import pytest

from harlequin.cli import DEFAULT_KEYMAP_NAMES
from harlequin.exception import HarlequinConfigError
from harlequin.keymap import HarlequinKeyMap
from harlequin.plugins import load_keymap_plugins


def test_load_keymap_plugins() -> None:
    built_in_keymaps = load_keymap_plugins(user_defined_keymaps=[])
    assert len(built_in_keymaps) == 1
    assert DEFAULT_KEYMAP_NAMES[0] in built_in_keymaps
    assert isinstance(built_in_keymaps[DEFAULT_KEYMAP_NAMES[0]], HarlequinKeyMap)
    assert built_in_keymaps[DEFAULT_KEYMAP_NAMES[0]].bindings


def test_do_not_load_keymaps_that_duplicate_plugin_names() -> None:
    my_map = HarlequinKeyMap(name=DEFAULT_KEYMAP_NAMES[0], bindings=[])
    with pytest.raises(HarlequinConfigError) as exc_info:
        _ = load_keymap_plugins(user_defined_keymaps=[my_map])
    assert "vscode" in str(exc_info)



from __future__ import annotations

from harlequin.autocomplete.completion import HarlequinCompletion

KEYWORDS = [
    "alter",
    "and",
    "as",
    "between",
    "by",
    "cascade",
    "case",
    "column",
    "copy",
    "create",
    "cross",
    "current",
    "database",
    "delete",
    "distinct",
    "drop",
    "end",
    "except",
    "exclude",
    "exists",
    "false",
    "filter",
    "following",
    "from",
    "full",
    "function",
    "grant",
    "group",
    "having",
    "if",
    "ilike",
    "inner",
    "insert",
    "intersect",
    "join",
    "lateral",
    "left",
    "like",
    "limit",
    "merge",
    "natural",
    "not",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "over",
    "owner",
    "partition",
    "preceding",
    "qualify",
    "range",
    "rename",
    "replace",
    "restrict",
    "revoke",
    "right",
    "row",
    "rows",
    "schema",
    "select",
    "sequence",
    "set",
    "similar",
    "table",
    "temp",
    "temporary",
    "then",
    "to",
    "top",
    "true",
    "truncate",
    "unbounded",
    "union",
    "update",
    "using",
    "view",
    "when",
    "where",
    "with",
]

FUNCTIONS = [
    ("abs", "fn"),
    ("ceil", "fn"),
    ("concat", "fn"),
    ("floor", "fn"),
    ("left", "fn"),
    ("lower", "fn"),
    ("ltrim", "fn"),
    ("regexp_extract", "fn"),
    ("regexp_replace", "fn"),
    ("replace", "fn"),
    ("right", "fn"),
    ("round", "fn"),
    ("rtrim", "fn"),
    ("sqrt", "fn"),
    ("avg", "agg"),
    ("bool_and", "agg"),
    ("bool_or", "agg"),
    ("count", "agg"),
    ("max", "agg"),
    ("min", "agg"),
    ("sum", "agg"),
]


def get_keywords() -> list[HarlequinCompletion]:
    """
    Returns a list of HarlequinCompletions.
    """
    keywords = [
        HarlequinCompletion(
            label=kw_name,
            type_label="kw",
            value=kw_name,
            priority=100,
            context=None,
        )
        for kw_name in KEYWORDS
    ]
    return keywords


def get_functions() -> list[HarlequinCompletion]:
    """
    Returns a list of HarlequinCompletions.
    """
    functions = [
        HarlequinCompletion(
            label=fn_name,
            type_label=fn_type,
            value=fn_name,
            priority=200,
            context=None,
        )
        for fn_name, fn_type in FUNCTIONS
    ]
    return functions



from __future__ import annotations

from pathlib import Path
from typing import Sequence, TypedDict, cast

from platformdirs import user_config_path
from tomlkit.exceptions import TOMLKitError
from tomlkit.toml_document import TOMLDocument
from tomlkit.toml_file import TOMLFile

from harlequin.exception import HarlequinConfigError
from harlequin.keymap import HarlequinKeyMap, RawKeyBinding


class Profile(TypedDict, total=False):
    conn_str: Sequence[str] | str
    adapter: str
    limit: str | int
    theme: str
    keymap_name: list[str]
    show_files: Path | str | None
    show_s3: str | None
    locale: str
    no_download_tzdata: bool
    # many more keys for adapter options


class Config(TypedDict, total=False):
    default_profile: str | None
    keymaps: dict[str, list[RawKeyBinding]]
    profiles: dict[str, Profile]


class ConfigFile:
    def __init__(self, path: Path) -> None:
        """
        Opens and reads the TOML file at path. Can be used to create
        a new file if one does not already exist.

        Stores references to the TOMLFile, TOMLDocument, and tracks
        whether or not the file is a pyproject.toml file.

        Raises: HarlequinConfigError if we can't read the TOML file.
        """
        self.path = path
        self.toml_file = TOMLFile(path)
        try:
            self.toml_doc = self.toml_file.read()
        except OSError:
            self.toml_doc = TOMLDocument()
        except TOMLKitError as e:
            raise HarlequinConfigError(
                f"Attempted to load the config file at {path}, but encountered an "
                f"error:\n\n{e}",
                title="Harlequin could not load the config file.",
            ) from e
        self.is_pyproject = path.stem == "pyproject"

    @property
    def relevant_config(self) -> Config:
        """
        Reads the relevant config section from a dedicated config file
        or pyproject.toml file at path. Raises HarlequinConfigError
        if there is a problem with the file.
        """
        relevant_config: Config = cast(
            Config,
            self.toml_doc.unwrap()
            if not self.is_pyproject
            else self.toml_doc.unwrap().get("tool", {}).get("harlequin", {}),
        )
        return relevant_config

    def update(self, config: Config) -> None:
        """
        Replace the relevant section of the in-memory TOML doc with the updated
        Config.
        """
        if self.is_pyproject:
            if "tool" not in self.toml_doc:
                self.toml_doc["tool"] = {"harlequin": {}}
            elif "harlequin" not in self.toml_doc["tool"]:  # type: ignore
                self.toml_doc["tool"]["harlequin"] = {}  # type: ignore
            self.toml_doc["tool"]["harlequin"].update(config)  # type: ignore
        else:
            self.toml_doc.update(config)

    def write(self) -> None:
        """
        Write the in-memory TOML doc to disk, at self.path.
        """
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.toml_file.write(self.toml_doc)


def get_config_for_profile(
    config_path: Path | None, profile_name: str | None
) -> tuple[Profile, list[HarlequinKeyMap]]:
    config = load_config(config_path)
    if not profile_name:
        profile_name = config.get("default_profile", None)

    if profile_name is None or profile_name == "None":
        profile: Profile = {}
    elif profile_name not in config.get("profiles", {}):
        raise HarlequinConfigError(
            f"Could not load the profile named {profile_name} because it does not "
            "exist in any discovered config files.",
            title="Harlequin couldn't load your profile.",
        )
    else:
        profile = config["profiles"][profile_name]

    raw_keymaps: dict[str, list[RawKeyBinding]] = config.get("keymaps", {})
    keymaps: list[HarlequinKeyMap] = [
        HarlequinKeyMap.from_config(name=name, bindings=bindings)
        for name, bindings in raw_keymaps.items()
    ]

    return profile, keymaps


def load_config(config_path: Path | None) -> Config:
    paths = _find_config_files(config_path)
    config = _merge_config_files(paths)
    _raise_on_bad_schema(config)
    return config


def get_highest_priority_existing_config_file() -> Path | None:
    """
    Returns the closest existing config file using the default search path;
    checks pyproject files for a tool.harlequin section and ignores those
    that are missing that section. Returns None if no
    config files are found.
    """
    candidates = _find_config_files(config_path=None)
    while candidates:
        p = candidates.pop()
        if p.stem == "pyproject":
            try:
                config_file = ConfigFile(p)
            except HarlequinConfigError:
                continue
            if not config_file.relevant_config:
                continue
        return p
    return None


def sluggify_option_name(raw: str) -> str:
    return raw.strip("-").replace("-", "_")


def _find_config_files(config_path: Path | None) -> list[Path]:
    """
    Returns a list of candidate config file paths, to be read and
    merged. Returns an empty list if none already exist. Order matters:
    the last item will have highest priority.
    """
    found_files: list[Path] = []
    for search in [_search_home, _search_config, _search_cwd]:
        found_files.extend(search())
    if config_path is not None and config_path.exists():
        found_files.append(config_path)
    elif config_path is not None:
        raise HarlequinConfigError(
            f"Config file could not be found at specified path: {config_path}",
            title="Harlequin couldn't load your config file.",
        )
    return found_files


def _search_cwd() -> list[Path]:
    directory = Path.cwd()
    filenames = ["pyproject.toml", ".harlequin.toml", "harlequin.toml"]
    return [directory / f for f in filenames if (directory / f).exists()]


def _search_config() -> list[Path]:
    directory = user_config_path(appname="harlequin", appauthor=False)
    filenames = ["config.toml", ".harlequin.toml", "harlequin.toml"]
    return [directory / f for f in filenames if (directory / f).exists()]


def _search_home() -> list[Path]:
    directory = Path.home()
    filenames = ["pyproject.toml", ".harlequin.toml", "harlequin.toml"]
    return [directory / f for f in filenames if (directory / f).exists()]


def _merge_config_files(paths: list[Path]) -> Config:
    config: Config = {}
    for p in paths:
        config_file = ConfigFile(p)
        config.update(config_file.relevant_config)
    return config


def _raise_on_bad_schema(config: Config) -> None:
    TOP_LEVEL_KEYS = ("default_profile", "profiles", "keymaps")
    if not config:
        return

    for k in config.keys():
        if k not in TOP_LEVEL_KEYS:
            raise HarlequinConfigError(
                f"Found unexpected key in config: {k}.\n"
                f"Allowed values are {TOP_LEVEL_KEYS}.",
                title="Harlequin couldn't load your config file.",
            )
    if config.get("profiles", None) is None:
        pass
    elif not isinstance(config["profiles"], dict):
        raise HarlequinConfigError(
            "The profiles key must define a table.",
            title="Harlequin couldn't load your config file.",
        )
    elif not all(
        [isinstance(config["profiles"][k], dict) for k in config["profiles"].keys()]
    ):
        raise HarlequinConfigError(
            "The members of the profiles table must be tables.",
            title="Harlequin couldn't load your config file.",
        )
    elif any(k == "None" for k in config["profiles"].keys()):
        raise HarlequinConfigError(
            "Config file defines a profile named 'None', which is not allowed.",
            title="Harlequin couldn't load your config file.",
        )
    else:
        for profile_name, opt_dict in config["profiles"].items():
            for option_name in opt_dict.keys():
                if "-" in option_name:
                    raise HarlequinConfigError(
                        f"Profile {profile_name} defines an option {option_name!r}, "
                        "which is an invalid name for an option. Did you mean "
                        f"{sluggify_option_name(option_name)!r}?",
                        title="Harlequin couldn't load your config file.",
                    )
                elif "keymap_names" in option_name:
                    raise HarlequinConfigError(
                        f"Profile {profile_name} defines an option {option_name!r}, "
                        "which is an invalid name for an option. Did you mean "
                        "'keymap_name' (singular)?",
                        title="Harlequin couldn't load your config file.",
                    )

    if config.get("keymaps", None) is None:
        pass
    elif not isinstance(config["keymaps"], dict):
        raise HarlequinConfigError(
            "The keymaps key must define a table.",
            title="Harlequin couldn't load your config file.",
        )
    elif not all(
        [isinstance(config["keymaps"][k], list) for k in config["keymaps"].keys()]
    ):
        raise HarlequinConfigError(
            "The members of each keymaps table must be arrays of tables.",
            title="Harlequin couldn't load your config file.",
        )

    if (default := config.get("default_profile", None)) is not None and not isinstance(
        default, str
    ):
        raise HarlequinConfigError(
            f"Config file sets default_profile to {default}, but that value "
            "must be a string.",
            title="Harlequin couldn't load your config file.",
        )
    elif (
        default is not None
        and isinstance(default, str)
        and isinstance(config["profiles"], dict)
        and default != "None"
        and config["profiles"].get(default, None) is None
    ):
        raise HarlequinConfigError(
            f"Config files set the default_profile to {default}, but do not define a "
            "profile with that name.",
            title="Harlequin couldn't load your config file.",
        )



from typing import Awaitable, Callable, List

import pytest

from harlequin import Harlequin


@pytest.mark.asyncio
async def test_toggle_sidebar(
    app: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    snap_results: List[bool] = []
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None or app.data_catalog.database_tree.loading:
            await pilot.pause()
        # initialization
        sidebar = app.data_catalog
        assert not sidebar.disabled
        assert sidebar.styles.width
        assert sidebar.styles.width.value > 0
        snap_results.append(await app_snapshot(app, "Initialization"))

        await pilot.press("ctrl+b")
        assert sidebar.disabled
        assert sidebar.styles.width
        assert sidebar.styles.width.value == 0
        snap_results.append(await app_snapshot(app, "Hidden"))

        await pilot.press("ctrl+b")
        assert not sidebar.disabled
        assert sidebar.styles.width
        assert sidebar.styles.width.value > 0
        snap_results.append(await app_snapshot(app, "Unhidden"))

        await pilot.press("f9")
        assert sidebar.disabled
        assert sidebar.styles.width
        assert sidebar.styles.width.value == 0
        snap_results.append(await app_snapshot(app, "Hidden Again"))

        assert all(snap_results)


@pytest.mark.asyncio
async def test_toggle_full_screen(
    app: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    snap_results: List[bool] = []
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        # initialization; all visible
        app.editor.focus()
        assert app.full_screen is False
        assert app.sidebar_hidden is False
        widgets = [app.data_catalog, app.editor_collection, app.results_viewer]
        for w in widgets:
            assert not w.disabled
            assert w.styles.width
            assert w.styles.width.value > 0
        snap_results.append(await app_snapshot(app, "Initialization"))

        await pilot.press("f10")
        # only editor visible
        assert not app.editor_collection.disabled
        assert not app.editor.disabled
        assert not app.run_query_bar.disabled
        assert app.editor_collection.styles.width
        assert app.editor_collection.styles.width.value > 0
        for w in [w for w in widgets if w != app.editor_collection]:
            assert w.disabled
            assert w.styles.width
            assert w.styles.width.value == 0
        snap_results.append(await app_snapshot(app, "Editor Full Screen"))

        await pilot.press("ctrl+b")
        # editor and data catalog should be visible
        assert not app.sidebar_hidden
        assert not app.data_catalog.disabled
        assert app.full_screen
        assert not app.editor_collection.disabled
        assert not app.editor.disabled
        snap_results.append(await app_snapshot(app, "Editor Full Screen with Sidebar"))

        await pilot.press("f10")
        # all visible
        for w in widgets:
            assert not w.disabled
            assert w.styles.width
            assert w.styles.width.value > 0
        snap_results.append(
            await app_snapshot(app, "Exit Full Screen (sidebar already visible)")
        )

        await pilot.press("ctrl+b")
        # data catalog hidden
        assert app.sidebar_hidden
        assert app.data_catalog.disabled
        assert not app.editor_collection.disabled
        assert not app.editor.disabled
        snap_results.append(await app_snapshot(app, "Sidebar hidden"))

        await pilot.press("f10")
        # only editor visible
        assert not app.editor_collection.disabled
        assert not app.editor.disabled
        assert app.data_catalog.disabled
        assert app.results_viewer.disabled
        snap_results.append(
            await app_snapshot(app, "Editor Full Screen (sidebar already hidden)")
        )

        await pilot.press("f10")
        # data catalog should still be hidden
        assert not app.editor_collection.disabled
        assert not app.editor.disabled
        assert not app.run_query_bar.disabled
        assert app.data_catalog.disabled
        assert not app.results_viewer.disabled
        snap_results.append(
            await app_snapshot(app, "Exit Full Screen (sidebar remains hidden)")
        )
        app.editor.text = "select 1"
        await pilot.press("ctrl+j")

        app.results_viewer.focus()
        await pilot.press("f10")
        # only results viewer should be visible
        assert app.editor_collection.disabled
        assert app.run_query_bar.disabled
        assert app.data_catalog.disabled
        assert not app.results_viewer.disabled
        snap_results.append(await app_snapshot(app, "Results Viewer Full Screen"))

        await pilot.press("f9")
        # results viewer and data catalog should be visible
        assert not app.sidebar_hidden
        assert not app.data_catalog.disabled
        assert app.full_screen
        assert app.editor_collection.disabled
        assert app.run_query_bar.disabled
        assert not app.results_viewer.disabled
        snap_results.append(
            await app_snapshot(app, "Results Viewer Full Screen with Sidebar")
        )

        await pilot.press("f10")
        # all visible
        assert not app.sidebar_hidden
        assert not app.full_screen
        for w in widgets:
            assert not w.disabled
            assert w.styles.width
            assert w.styles.width.value > 0
        snap_results.append(
            await app_snapshot(app, "Exit RV Full Screen (sidebar visible)")
        )

        assert all(snap_results)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, Sequence, Tuple

from textual import events
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.css.query import QueryError
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, Static
from textual_textarea import PathInput

from harlequin.components.error_modal import ErrorModal
from harlequin.components.results_viewer import ResultsTable
from harlequin.exception import HarlequinCopyError
from harlequin.export import copy
from harlequin.options import AbstractOption, HarlequinCopyFormat

ExportOptions = Dict[str, Any]


def export_callback(
    screen_data: Tuple[Path, str, ExportOptions],
    table: ResultsTable,
    success_callback: Callable[[], None],
    error_callback: Callable[[Exception], None],
) -> None:
    try:
        copy(
            table=table,
            path=screen_data[0],
            format_name=screen_data[1],
            options=screen_data[2],
        )
        success_callback()
    except (OSError, HarlequinCopyError) as e:
        error_callback(e)


class NoFocusVerticalScroll(VerticalScroll, can_focus=False):
    pass


class CopyOptionsMenu(Widget, can_focus=False):
    def __init__(
        self,
        format_name: str,
        options: Sequence[AbstractOption],
        *children: Widget,
        name: str | None = None,
        id: str | None = None,  # noqa: A002
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )
        self.format_name = format_name
        self.options = options

    def compose(self) -> ComposeResult:
        for option in self.options:
            with Horizontal(classes="option_row"):
                yield from option.to_widgets()

    def on_mount(self) -> None:
        for option in self.options:
            w = self._get_option_widget_by_name(option.name)
            if w is not None:
                w.tooltip = option.description

    @property
    def current_options(self) -> ExportOptions:
        return {
            option.name: self._get_option_value_by_name(option.name)
            for option in self.options
        }

    def _get_option_widget_by_name(self, name: str) -> Widget | None:
        try:
            return self.query_one(f"#{name}")
        except QueryError:
            return None

    def _get_option_value_by_name(self, name: str) -> Any | None:
        w = self._get_option_widget_by_name(name)
        if w:
            return getattr(w, "value", None)
        else:
            return None


class ExportScreen(ModalScreen[Tuple[Path, str, ExportOptions]]):
    def __init__(
        self,
        formats: list[HarlequinCopyFormat],
        name: str | None = None,
        id: str | None = None,  # noqa: A002
        classes: str | None = None,
    ) -> None:
        super().__init__(name, id, classes)
        self.formats = formats

    def compose(self) -> ComposeResult:
        assert self.formats is not None
        with Vertical(id="export_outer"):
            yield Static(
                "Export the results of your query to a file.",
                id="export_header",
            )
            yield PathInput(
                placeholder=(
                    "/path/to/file  (tab autocompletes, enter exports, esc cancels)"
                ),
                id="path_input",
                file_okay=True,
                dir_okay=False,
                must_exist=False,
                tab_advances_focus=True,
            )
            yield Label("", id="validation_label")
            with Horizontal(classes="option_row"):
                yield Label("Format:", classes="select_label")
                yield Select(
                    options=[(option.label, option.name) for option in self.formats],
                    id="format_select",
                )
            yield NoFocusVerticalScroll(id="options_container")
            with Horizontal(id="export_button_row"):
                yield Button(label="Cancel", variant="error", id="cancel")
                yield Button(label="Export", variant="primary", id="export")

    def on_mount(self) -> None:
        container = self.query_one("#export_outer")
        container.border_title = "Data Exporter"
        self.format_select = self.query_one(Select)
        self.file_input = self.query_one("#path_input", Input)
        self.file_input_validation_label = self.query_one("#validation_label", Label)
        self.options_container = self.query_one(
            "#options_container", NoFocusVerticalScroll
        )
        self.export_button = self.query_one("#export", Button)
        self.file_input.focus()

    def on_key(self, event: events.Key) -> None:
        if event.key == "escape":
            self.app.pop_screen()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        button = event.button
        if button.id == "export":
            self._export()
        else:
            self.app.pop_screen()

    def on_input_changed(self, event: Input.Changed) -> None:
        event.stop()
        if event.control.id == "path_input":
            if event.validation_result:
                if event.validation_result.is_valid:
                    self.file_input_validation_label.update("")
                else:
                    self.file_input_validation_label.update(
                        " ".join(event.validation_result.failure_descriptions)
                    )
            old_format = self.format_select.value
            new_format = self._get_format_from_file_extension(event.value)
            if new_format:
                self.format_select.value = new_format
            if self.format_select.value != old_format:
                self.post_message(
                    Select.Changed(self.format_select, value=self.format_select.value)
                )

    def on_input_submitted(self, event: Input.Submitted) -> None:
        event.stop()
        if event.control.id == "path_input":
            self.export_button.press()
        else:
            self.focus_next()

    async def on_select_changed(self, event: Select.Changed) -> None:
        event.stop()
        if event.control.id == "format_select":
            for child in self.options_container.children:
                await child.remove()
            try:
                assert self.formats is not None
                [options] = [
                    fmt.options for fmt in self.formats if fmt.name == event.value
                ]
            except (ValueError, IndexError, AssertionError):
                return
            menu = CopyOptionsMenu(str(event.value), options)
            await self.options_container.mount(menu)

    def _export(self) -> None:
        path = Path(self.file_input.value)
        if path.is_dir():
            self.app.push_screen(
                ErrorModal(
                    title="Error Writing File",
                    header="Path is not a file",
                    error=OSError(f"Cannot write to {path}, since it is a directory."),
                )
            )
        elif self.format_select.value is None:
            self.app.push_screen(
                ErrorModal(
                    title="Error Writing File",
                    header="Must select format",
                    error=OSError(
                        "You must select a file format "
                        f"{[fmt.label for fmt in self.formats]}"
                    ),
                )
            )
        else:
            try:
                options_menu = self.query_one(CopyOptionsMenu)
            except QueryError:
                return
            else:
                self.dismiss(
                    (path, options_menu.format_name, options_menu.current_options)
                )

    def _get_format_from_file_extension(self, input_value: str) -> str | None:
        mapping = {ext: fmt.name for fmt in self.formats for ext in fmt.extensions}
        try:
            p = Path(input_value)
            format_name = mapping[p.suffix]
        except (ValueError, KeyError):
            return None
        else:
            return format_name



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from textual.message import Message

if TYPE_CHECKING:
    from textual.notifications import SeverityLevel

    from harlequin.app import Harlequin


class HarlequinDriver:
    """
    Provides an interface for Data Catalog interactions to drive
    the Harlequin UI. These will always be called from worker
    threads, so they use Messages to update the main Harlequin
    state.
    """

    class ConfirmAndExecute(Message):
        def __init__(self, callback: Callable[[], None], instructions: str) -> None:
            super().__init__()
            self.callback = callback
            self.instructions = instructions

    class InsertTextAtSelection(Message):
        def __init__(self, text: str) -> None:
            super().__init__()
            self.text = text

    class InsertTextInNewBuffer(Message):
        def __init__(self, text: str) -> None:
            super().__init__()
            self.text = text

    class Notify(Message):
        def __init__(self, notify_message: str, severity: "SeverityLevel") -> None:
            super().__init__()
            self.notify_message = notify_message
            self.severity = severity

    class Refreshcatalog(Message):
        pass

    def __init__(self, app: Harlequin) -> None:
        self.app = app

    def insert_text_at_selection(self, text: str) -> None:
        """
        Insert text at the cursor in Harlequin's editor.
        """
        self.app.post_message(self.InsertTextAtSelection(text=text))

    def insert_text_in_new_buffer(self, text: str) -> None:
        """
        Create a new buffer (tab) in Harlequin's editor and insert the text.
        """
        self.app.post_message(self.InsertTextInNewBuffer(text=text))

    def confirm_and_execute(
        self, callback: Callable[[], None], instructions: str = "Are you sure?"
    ) -> None:
        """
        Display a confirmation modal in Harlequin; If the user selects "Yes",
        then Harlequin will invoke the callback.
        """
        self.app.post_message(
            self.ConfirmAndExecute(callback=callback, instructions=instructions)
        )

    def notify(self, message: str, severity: "SeverityLevel" = "information") -> None:
        """
        Show a toast notification in Harlequin.
        """
        self.app.post_message(self.Notify(notify_message=message, severity=severity))

    def refresh_catalog(self) -> None:
        """
        Force a refresh of the catalog.
        """
        self.app.post_message(self.Refreshcatalog())



from __future__ import annotations

from pathlib import Path
from sqlite3 import Connection

from harlequin.options import (
    FlagOption,
    ListOption,
    PathOption,
    SelectOption,
    TextOption,
)

read_only = FlagOption(
    name="read-only",
    short_decls=["-readonly", "-r"],
    description="Open the database file in read-only mode.",
)
connection_mode = SelectOption(
    name="mode",
    short_decls=["-mode", "-m"],
    description=(
        "The mode parameter may be set to either 'ro', 'rw', 'rwc', or 'memory'. "
        "Attempting to set it to any other value is an error. If 'ro' is specified, "
        "then the database is opened for read-only access. If the mode option is set "
        "to 'rw', then the database is opened for read-write (but not create) access. "
        "Value 'rwc' is equivalent to setting both SQLITE_OPEN_READWRITE and "
        "SQLITE_OPEN_CREATE. If the mode option is set to 'memory' then a pure "
        "in-memory database that never reads or writes from disk is used. "
        "It is an error to specify a value for the mode parameter that is less "
        "restrictive than the separate --read-only option."
    ),
    choices=["ro", "rw", "rwc", "memory"],
    default="rwc",
)


def _float_validator(s: str | None) -> tuple[bool, str]:
    if s is None:
        return True, ""
    try:
        _ = float(s)
    except ValueError:
        return False, f"Cannot convert {s} to a float!"
    else:
        return True, ""


timeout = TextOption(
    name="timeout",
    description=(
        "How many seconds the connection should wait before raising an "
        "OperationalError when a table is locked. If another connection opens a "
        "transaction to modify a table, that table will be locked until the "
        "transaction is committed. Default five seconds."
    ),
    validator=_float_validator,
)


def _int_validator(s: str | None) -> tuple[bool, str]:
    if s is None:
        return True, ""
    try:
        _ = int(s)
    except ValueError:
        return False, f"Cannot convert {s} to an int!"
    else:
        return True, ""


detect_types = TextOption(
    name="detect-types",
    description=(
        "Control whether and how data types not natively supported by SQLite are "
        "looked up to be converted to Python types, using the converters registered "
        "with register_converter(). Set it to any combination (using |, bitwise or) "
        "of PARSE_DECLTYPES and PARSE_COLNAMES to enable this. Column names takes "
        "precedence over declared types if both flags are set. Types cannot be "
        "detected for generated fields (for example max(data)), even when the "
        "detect_types parameter is set; str will be returned instead. By default (0), "
        "type detection is disabled."
    ),
    validator=_int_validator,
)

isolation_level = SelectOption(
    name="isolation-level",
    description=(
        "Control legacy transaction handling behaviour. See Connection.isolation_level "
        "and Transaction control via the isolation_level attribute for more "
        'information. Can be "DEFERRED" (default), "EXCLUSIVE" or "IMMEDIATE".'
    ),
    choices=["DEFERRED", "EXCLUSIVE", "IMMEDIATE"],
    default="DEFERRED",
)


cached_statements = TextOption(
    name="cached-statements",
    description=(
        "The number of statements that sqlite3 should internally cache for this "
        "connection, to avoid parsing overhead. By default, 128 statements."
    ),
    validator=_int_validator,
)


init = PathOption(
    name="init-path",
    description=(
        "The path to an initialization script. On startup, Harlequin will execute "
        "the commands in the script against the attached database."
    ),
    short_decls=["-i", "-init"],
    exists=False,
    file_okay=True,
    dir_okay=False,
    resolve_path=True,
    path_type=Path,
)

no_init = FlagOption(
    name="no-init",
    description="Start Harlequin without executing the initialization script.",
)

extensions = ListOption(
    name="extension",
    description=(
        "Load the SQLite extension from the passed path when starting "
        "Harlequin. To install multiple extensions, repeat this option."
    ),
    short_decls=["-e"],
)

SQLITE_OPTIONS = [
    init,
    no_init,
    read_only,
    connection_mode,
    timeout,
    detect_types,
    cached_statements,
]

if hasattr(Connection, "enable_load_extension"):
    SQLITE_OPTIONS.append(extensions)

# Python 3.12 and lower
if not hasattr(Connection, "autocommit"):
    SQLITE_OPTIONS.append(isolation_level)



class ArrowException(Exception): ...
class ArrowInvalid(ValueError, ArrowException): ...
class ArrowMemoryError(MemoryError, ArrowException): ...
class ArrowKeyError(KeyError, Exception): ...
class ArrowTypeError(TypeError, Exception): ...
class ArrowNotImplementedError(NotImplementedError, ArrowException): ...
class ArrowCapacityError(ArrowException): ...
class ArrowIndexError(IndexError, ArrowException): ...
class ArrowSerializationError(ArrowException): ...
class ArrowCancelled(ArrowException): ...

ArrowIOError = IOError



from __future__ import annotations

import csv
import sqlite3
import sys
from pathlib import Path

import duckdb
import pytest

from harlequin import Harlequin
from harlequin.adapter import HarlequinAdapter
from harlequin.locale_manager import set_locale
from harlequin.windows_timezone import check_and_install_tzdata

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


@pytest.fixture(scope="session", autouse=True)
def install_tzdata() -> None:
    if sys.platform == "win32":
        check_and_install_tzdata()


@pytest.fixture(scope="session", autouse=True)
def set_locale_to_enUS() -> None:
    set_locale("en_US.UTF-8")


@pytest.fixture
def data_dir() -> Path:
    here = Path(__file__)
    return here.parent / "data"


@pytest.fixture
def tiny_duck(tmp_path: Path, data_dir: Path) -> Path:
    """
    Creates a duckdb database file from the contents of
    data_dir/functional_tests/tiny
    """
    path_to_data = data_dir / "functional_tests" / "tiny"
    path_to_db = tmp_path / "tiny.db"
    conn = duckdb.connect(str(path_to_db))
    conn.execute(f"import database '{path_to_data}';")
    return path_to_db


@pytest.fixture
def tiny_sqlite(tmp_path: Path, data_dir: Path) -> Path:
    path_to_data = data_dir / "functional_tests" / "tiny"
    path_to_db = tmp_path / "tiny.sqlite"
    _create_sqlite_db_from_data_dir(path_to_data, path_to_db)
    return path_to_db


@pytest.fixture
def small_duck(tmp_path: Path, data_dir: Path) -> Path:
    """
    Creates a duckdb database file from the contents of
    data_dir/functional_tests/small
    """
    path_to_data = data_dir / "functional_tests" / "small"
    path_to_db = tmp_path / "small.db"
    conn = duckdb.connect(str(path_to_db))
    conn.execute(f"import database '{path_to_data}';")
    return path_to_db


@pytest.fixture
def small_sqlite(tmp_path: Path, data_dir: Path) -> Path:
    path_to_data = data_dir / "functional_tests" / "small"
    path_to_db = tmp_path / "small.sqlite"
    _create_sqlite_db_from_data_dir(path_to_data, path_to_db)
    return path_to_db


@pytest.fixture
def duckdb_adapter() -> type[HarlequinAdapter]:
    eps = entry_points(group="harlequin.adapter")
    cls: type[HarlequinAdapter] = eps["duckdb"].load()
    return cls


@pytest.fixture
def sqlite_adapter() -> type[HarlequinAdapter]:
    eps = entry_points(group="harlequin.adapter")
    cls: type[HarlequinAdapter] = eps["sqlite"].load()
    return cls


@pytest.fixture(params=["duckdb", "sqlite"])
def all_adapters(request: pytest.FixtureRequest) -> type[HarlequinAdapter]:
    eps = entry_points(group="harlequin.adapter")
    cls: type[HarlequinAdapter] = eps[request.param].load()
    return cls


@pytest.fixture
def app(duckdb_adapter: type[HarlequinAdapter]) -> Harlequin:
    return Harlequin(duckdb_adapter([":memory:"], no_init=True), connection_hash="foo")


@pytest.fixture
def app_all_adapters(all_adapters: type[HarlequinAdapter]) -> Harlequin:
    return Harlequin(all_adapters([":memory:"], no_init=True), connection_hash="foo")


@pytest.fixture
def app_small_duck(
    duckdb_adapter: type[HarlequinAdapter], small_duck: Path
) -> Harlequin:
    return Harlequin(
        duckdb_adapter([str(small_duck)], no_init=True), connection_hash="small"
    )


@pytest.fixture
def app_small_sqlite(
    sqlite_adapter: type[HarlequinAdapter], small_sqlite: Path
) -> Harlequin:
    return Harlequin(
        sqlite_adapter([str(small_sqlite)], no_init=True), connection_hash="bar"
    )


@pytest.fixture(params=["duckdb", "sqlite"])
def app_all_adapters_small_db(
    request: pytest.FixtureRequest,
    app_small_duck: Harlequin,
    app_small_sqlite: Harlequin,
) -> Harlequin:
    if request.param == "duckdb":
        return app_small_duck
    else:
        return app_small_sqlite


@pytest.fixture
def app_multi_duck(
    duckdb_adapter: type[HarlequinAdapter], tiny_duck: Path, small_duck: Path
) -> Harlequin:
    return Harlequin(
        duckdb_adapter([str(tiny_duck), str(small_duck)], no_init=True),
        connection_hash="multi",
    )


def _create_sqlite_db_from_data_dir(data_dir: Path, db_path: Path) -> None:
    SINGLE_QUOTE = "'"
    DOUBLED_SINGLE_QUOTE = "''"
    conn = sqlite3.connect(str(db_path))
    with open(data_dir / "schema_sqlite.sql") as f:
        ddl = f.read()
    for q in ddl.split(";"):
        conn.execute(q)
    for p in data_dir.iterdir():
        if p.is_file() and p.suffix == ".csv":
            with p.open("r", encoding="utf-8") as f:
                reader = csv.reader(f)
                for row in reader:
                    quoted = [
                        (
                            val
                            if isinstance(val, (int, float))
                            else f"'{val.replace(SINGLE_QUOTE, DOUBLED_SINGLE_QUOTE)}'"
                        )
                        for val in row
                    ]
                    conn.execute(f"insert into {p.stem} values({', '.join(quoted)})")
                conn.commit()



from . import Table

def write_feather(
    df: Table,
    dest: str,
    compression: str | None = None,
    compression_level: int | None = None,
    chunksize: int | None = None,
    version: int = 2,
) -> None: ...



from __future__ import annotations

import sys
from typing import Awaitable, Callable, List

import pytest
from textual.message import Message
from textual.notifications import Notify
from textual.widgets.text_area import Selection
from textual.widgets.tree import TreeNode

from harlequin import Harlequin
from harlequin.app import QuerySubmitted
from harlequin.catalog import CatalogItem


def transaction_button_visible(app: Harlequin) -> bool:
    """
    Skip snapshot checks for versions of that app showing the autocommit button.
    """
    return sys.version_info >= (3, 12) and "Sqlite" in app.adapter.__class__.__name__


@pytest.mark.asyncio
async def test_query_formatting(
    app: Harlequin,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = "select\n\n1 FROM\n\n foo"

        await pilot.press("f4")
        assert app.editor.text == "select 1 from foo\n"


@pytest.mark.asyncio
async def test_multiple_buffers(
    app: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    snap_results: List[bool] = []
    async with app.run_test(size=(120, 36)) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        assert app.editor_collection
        assert app.editor_collection.tab_count == 1
        assert app.editor_collection.active == "tab-1"
        app.editor.text = "tab 1"
        await pilot.press("home")
        await pilot.pause()
        snap_results.append(await app_snapshot(app, "Tab 1 of 1 (No tabs)"))

        await pilot.press("ctrl+n")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.tab_count == 2
        assert app.editor_collection.active == "tab-2"
        assert app.editor.text == ""
        app.editor.text = "tab 2"
        await pilot.press("home")
        await pilot.pause()
        snap_results.append(await app_snapshot(app, "Tab 2 of 2"))

        await pilot.press("ctrl+n")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.tab_count == 3
        assert app.editor_collection.active == "tab-3"
        assert app.editor.text == ""
        app.editor.text = "tab 3"
        await pilot.press("home")
        await pilot.pause()
        snap_results.append(await app_snapshot(app, "Tab 3 of 3"))

        await pilot.press("ctrl+k")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.tab_count == 3
        assert app.editor_collection.active == "tab-1"
        assert app.editor.text == "tab 1"
        snap_results.append(await app_snapshot(app, "Tab 1 of 3"))

        await pilot.press("ctrl+k")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.tab_count == 3
        assert app.editor_collection.active == "tab-2"
        assert app.editor.text == "tab 2"
        snap_results.append(await app_snapshot(app, "Tab 2 of 3"))

        await pilot.press("ctrl+w")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.tab_count == 2
        assert app.editor_collection.active == "tab-3"
        assert app.editor.text == "tab 3"
        # TODO: bring back this flaky test.
        # snap_results.append(await app_snapshot(app, "Tab 3 after deleting 2"))

        await pilot.press("ctrl+k")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.active == "tab-1"
        assert app.editor.text == "tab 1"
        snap_results.append(await app_snapshot(app, "Tab 1 of [1,3]"))

        await pilot.press("ctrl+k")
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        assert app.editor_collection.active == "tab-3"
        assert app.editor.text == "tab 3"
        snap_results.append(await app_snapshot(app, "Tab 3 of [1,3]"))

        assert all(snap_results)


@pytest.mark.asyncio
async def test_word_autocomplete(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    snap_results: List[bool] = []
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None or app.editor_collection.word_completer is None:
            await pilot.pause()

        # we need to let the data catalog load the root's children
        while (
            app.data_catalog.database_tree.loading
            or not app.data_catalog.database_tree.root.children
        ):
            await pilot.pause()

        app.editor.focus()

        await pilot.press("s")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "s"))

        await pilot.press("e")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "se"))

        await pilot.press("l")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "sel"))

        await pilot.press("backspace")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "se again"))

        await pilot.press("l")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        await pilot.press("enter")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "submitted"))

        if not (transaction_button_visible(app)):
            assert all(snap_results)


@pytest.mark.skipif(
    sys.platform == "win32", reason="Initial snapshot very flaky on windows."
)
@pytest.mark.asyncio
async def test_member_autocomplete(
    app_small_duck: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_small_duck
    snap_results: List[bool] = []
    async with app.run_test() as pilot:
        await wait_for_workers(app)

        async def _expand_and_wait(node: TreeNode[CatalogItem]) -> None:
            node.expand()
            while not node.children:
                if getattr(node.data, "loaded", True):
                    break
                await pilot.pause()

        # we need to expand the data catalog to load items into the completer
        while (
            app.data_catalog.database_tree.loading
            or not app.data_catalog.database_tree.root.children
        ):
            await pilot.pause()
        for db_node in app.data_catalog.database_tree.root.children:
            await _expand_and_wait(db_node)
            await wait_for_workers(app)
            for schema_node in db_node.children:
                await _expand_and_wait(schema_node)
                await wait_for_workers(app)
        await pilot.pause(1)

        # now the completer should be populated
        while app.editor is None or app.editor_collection.member_completer is None:
            await pilot.pause()

        app.editor.text = '"drivers"'
        app.editor.selection = Selection((0, 9), (0, 9))
        app.editor.focus()

        await pilot.press("full_stop")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "driver members"))

        await pilot.press("quotation_mark")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "with quote"))

        await pilot.press("enter")
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "submitted"))

        assert all(snap_results)


@pytest.mark.asyncio
async def test_no_tree_sitter(
    app: Harlequin,
    monkeypatch: pytest.MonkeyPatch,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    import textual.document._syntax_aware_document
    import textual.widgets._text_area

    monkeypatch.setattr(textual.document._syntax_aware_document, "TREE_SITTER", False)
    monkeypatch.setattr(textual.widgets._text_area, "TREE_SITTER", False)

    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)

        while app.editor is None:
            await pilot.pause()

        assert app.editor is not None
        assert app.editor.text_input is not None
        app.editor.text = "select 1; select 2"

        assert not app.editor.text_input.is_syntax_aware

        await pilot.press("ctrl+a")
        await pilot.press("ctrl+j")
        await pilot.pause()
        await wait_for_workers(app)

        submitted_msg = next(
            iter(filter(lambda m: isinstance(m, QuerySubmitted), messages))
        )
        assert submitted_msg
        assert isinstance(submitted_msg, QuerySubmitted)
        assert len(submitted_msg.queries) == 2

        text_area_warning = next(
            iter(
                filter(
                    lambda m: isinstance(m, Notify)
                    and m.notification.severity == "warning",
                    messages,
                )
            )
        )
        assert text_area_warning



from __future__ import annotations

from datetime import datetime
from typing import Any, Callable, Literal

from . import DataType, MemoryPool, Scalar, _PandasConvertible

class Expression: ...
class ScalarAggregateOptions: ...

class CastOptions:
    def __init__(
        self,
        target_type: DataType | None = None,
        allow_int_overflow: bool | None = None,
        allow_time_truncate: bool | None = None,
        allow_time_overflow: bool | None = None,
        allow_decimal_truncate: bool | None = None,
        allow_float_truncate: bool | None = None,
        allow_invalid_utf8: bool | None = None,
    ) -> None: ...

def max(  # noqa: A001
    array: _PandasConvertible,
    /,
    *,
    skip_nulls: bool = True,
    min_count: int = 1,
    options: ScalarAggregateOptions | None = None,
    memory_pool: MemoryPool | None = None,
) -> Scalar: ...
def utf8_length(
    strings: _PandasConvertible, /, *, memory_pool: MemoryPool | None = None
) -> _PandasConvertible: ...
def register_scalar_function(
    func: Callable,
    function_name: str,
    function_doc: dict[Literal["summary", "description"], str],
    in_types: dict[str, DataType],
    out_type: DataType,
    func_registry: Any | None = None,
) -> None: ...
def call_function(
    function_name: str, target: list[_PandasConvertible]
) -> _PandasConvertible: ...
def assume_timezone(
    timestamps: _PandasConvertible | Scalar | datetime,
    /,
    timezone: str,
    *,
    ambiguous: Literal["raise", "earliest", "latest"] = "raise",
    nonexistent: Literal["raise", "earliest", "latest"] = "raise",
    options: Any | None = None,
    memory_pool: MemoryPool | None = None,
) -> _PandasConvertible: ...



from contextlib import suppress
from typing import Awaitable, Callable
from unittest.mock import MagicMock

import pytest
from textual.worker import WorkerCancelled

from harlequin.app import Harlequin
from harlequin.autocomplete import HarlequinCompletion


@pytest.fixture(autouse=True)
def no_use_buffer_cache(
    monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest
) -> None:
    if "use_cache" in request.keywords:
        return
    monkeypatch.setattr("harlequin.components.code_editor.load_cache", lambda: None)
    monkeypatch.setattr("harlequin.app.write_editor_cache", lambda *_: None)


@pytest.fixture(autouse=True)
def no_use_catalog_cache(
    monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest
) -> None:
    if "use_cache" in request.keywords:
        return
    monkeypatch.setattr("harlequin.app.get_catalog_cache", lambda *_: None)
    monkeypatch.setattr("harlequin.app.update_catalog_cache", lambda *_: None)


@pytest.fixture(autouse=True)
def mock_config_loader(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "harlequin.cli.get_config_for_profile", lambda **_: (dict(), [])
    )


@pytest.fixture(autouse=True)
def mock_completions(monkeypatch: pytest.MonkeyPatch) -> None:
    KEYWORDS = [
        "abort",
        "all",
        "alter",
        "always",
        "analyze",
        "and",
        "as",
        "asc",
        "begin",
        "between",
        "by",
        "cascade",
        "case",
        "column",
        "commit",
        "create",
        "database",
        "delete",
        "desc",
        "distinct",
        "drop",
        "else",
        "end",
        "explain",
        "from",
        "group",
        "groups",
        "having",
        "in",
        "inner",
        "insert",
        "intersect",
        "into",
        "is",
        "join",
        "left",
        "like",
        "limit",
        "null",
        "on",
        "order",
        "outer",
        "over",
        "partition",
        "row",
        "savepoint",
        "select",
        "set",
        "table",
        "temp",
        "temporary",
        "then",
        "union",
        "update",
        "using",
        "values",
        "view",
        "when",
        "where",
        "window",
    ]

    FUNCTIONS = [
        ("array_select", "fn"),
        ("count", "agg"),
        ("greatest", "fn"),
        ("least", "fn"),
        ("list_select", "fn"),
        ("sqrt", "fn"),
        ("sum", "agg"),
    ]

    keyword_completions = [
        HarlequinCompletion(
            label=kw_name, type_label="kw", value=kw_name, priority=100, context=None
        )
        for kw_name in KEYWORDS
    ]

    function_completions = [
        HarlequinCompletion(
            label=label, type_label=type_label, value=label, priority=1000, context=None
        )
        for label, type_label in FUNCTIONS
    ]

    completions = [*keyword_completions, *function_completions]
    duckdb_completions = [
        (
            completion.label,
            completion.type_label,
            completion.priority,
            completion.context,
        )
        for completion in completions
    ]
    monkeypatch.setattr(
        "harlequin_sqlite.adapter.get_completion_data",
        lambda *_: completions,
        raising=True,
    )
    monkeypatch.setattr(
        "harlequin_duckdb.adapter.get_completion_data",
        lambda *_: duckdb_completions,
        raising=True,
    )


@pytest.fixture
def mock_pyperclip(monkeypatch: pytest.MonkeyPatch) -> MagicMock:
    mock = MagicMock()
    mock.determine_clipboard.return_value = mock.copy, mock.paste

    def set_paste(x: str) -> None:
        mock.paste.return_value = x

    mock.copy.side_effect = set_paste
    monkeypatch.setattr("textual_textarea.text_editor.pyperclip", mock)

    return mock


@pytest.fixture
def wait_for_workers() -> Callable[[Harlequin], Awaitable[None]]:
    async def wait_for_filtered_workers(app: Harlequin) -> None:
        filtered_workers = [
            w for w in app.workers if w.name != "_database_tree_background_loader"
        ]
        if filtered_workers:
            with suppress(WorkerCancelled):
                await app.workers.wait_for_complete(filtered_workers)

    return wait_for_filtered_workers



from __future__ import annotations

from pathlib import Path
from typing import Any, Sequence, Tuple, Union

from platformdirs import user_config_path
from rich.panel import Panel
from rich.text import Text
from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.coordinate import Coordinate
from textual.driver import Driver
from textual.events import Key
from textual.message import Message
from textual.screen import ModalScreen
from textual.validation import ValidationResult, Validator
from textual.widget import Widget
from textual.widgets import Button, Footer, Input, Label, LoadingIndicator, Static
from textual_fastdatatable import DataTable

from harlequin.actions import HARLEQUIN_ACTIONS
from harlequin.app_base import AppBase
from harlequin.colors import YELLOW
from harlequin.config import (
    ConfigFile,
    get_config_for_profile,
    get_highest_priority_existing_config_file,
)
from harlequin.copy_widgets import NoFocusLabel, PathInput
from harlequin.exception import HarlequinError, pretty_error_message
from harlequin.keymap import HarlequinKeyBinding
from harlequin.plugins import load_keymap_plugins


class ActiveKeymapFound(Message):
    def __init__(self, names: list[str]) -> None:
        super().__init__()
        self.keymap_names = names


class PluginKeymapsFound(Message):
    def __init__(self, names: list[str]) -> None:
        super().__init__()
        self.keymap_names = names


class BindingsReady(Message):
    def __init__(
        self,
        bindings: dict[str, HarlequinKeyBinding],
        table_data: list[tuple[str, str, str]],
    ) -> None:
        super().__init__()
        self.bindings = bindings
        self.table_data = table_data


class EditButton(Button, inherit_bindings=False):
    BINDINGS = [Binding("enter", "press", "Edit Binding", show=True)]

    def __init__(self, key: str, classes: str | None = None):
        super().__init__(label=key, classes=classes)
        self.key = key


class RemoveButton(Button, inherit_bindings=False):
    BINDINGS = [Binding("enter", "press", "Remove Binding", show=True)]

    def __init__(self, key: str, classes: str | None = None):
        super().__init__(label="X", classes=classes)
        self.key = key


class AddButton(Button, inherit_bindings=False):
    BINDINGS = [Binding("enter", "press", "Add Binding", show=True)]


class BindingTable(DataTable, inherit_bindings=False):
    BINDINGS = [
        Binding("enter", "select_cursor", "Edit", show=True),
        Binding("up", "cursor_up", "Cursor Up", show=False),
        Binding("down", "cursor_down", "Cursor Down", show=False),
        Binding("ctrl+up", "scroll_home", "Home", show=False),
        Binding("ctrl+down", "scroll_end", "End", show=False),
        Binding("pageup", "page_up", "Page Up", show=False),
        Binding("pagedown", "page_down", "Page Down", show=False),
        Binding("home", "scroll_home", "Home", show=False),
        Binding("end", "scroll_end", "End", show=False),
        Binding("ctrl+home", "cursor_table_start", "Home", show=False),
        Binding("ctrl+end", "cursor_table_end", "End", show=False),
    ]

    def __init__(
        self,
        *,
        id: str | None = None,  # noqa: A002
        column_labels: list[str | Text] | None = None,
        data: Any | None = None,
    ) -> None:
        super().__init__(
            id=id,
            column_labels=column_labels,
            data=data,
            cursor_type="row",
        )


class KeymapNameValidator(Validator):
    def __init__(
        self,
        plugin_names: Sequence[str],
    ) -> None:
        super().__init__(
            failure_description="Cannot use the name of an existing keymap plug-in"
        )
        self.plugin_names = set(plugin_names)

    def validate(self, value: str) -> ValidationResult:
        if value in self.plugin_names:
            return self.failure()
        else:
            return self.success()


class QuitModal(ModalScreen[Tuple[bool, Union[Path, None], Union[str, None]]]):
    BINDINGS = [
        Binding("escape", "cancel", "Cancel"),
        Binding("up,left", "focus_previous", "focus_previous", show=False),
        Binding("down,right", "focus_next", "focus_next", show=False),
    ]

    def __init__(
        self, config_path: Path | None, keymap_name: str | None, plugin_names: list[str]
    ) -> None:
        super().__init__()
        self.config_path = (
            config_path
            or get_highest_priority_existing_config_file()
            or user_config_path(appname="harlequin", appauthor=False) / "config.toml"
        )
        self.keymap_name = keymap_name
        self.plugin_names = plugin_names
        self.path_input_valid = False
        self.name_input_valid = False

    def compose(self) -> ComposeResult:
        self.path_input = PathInput(
            id="path_input",
            value=str(self.config_path),
            dir_okay=False,
            tab_advances_focus=True,
        )
        self.path_validation_label = Label(
            "", id="path_validation_label", classes="validation-label"
        )
        name_validator = KeymapNameValidator(plugin_names=self.plugin_names)
        validation_result = name_validator.validate(self.keymap_name or "")
        self.name_input = Input(
            id="name_input",
            value=self.keymap_name if validation_result.is_valid else None,
            placeholder="Enter a name for this keymap",
            validators=[name_validator],
        )
        self.name_validation_label = Label(
            "", id="name_validation_label", classes="validation-label"
        )
        self.save_button = Button(label="Save + Quit", variant="primary", id="submit")
        with Vertical(id="outer"):
            with Horizontal(classes="option_row"):
                yield NoFocusLabel("Config Path:")
                with Vertical():
                    yield self.path_input
                    yield self.path_validation_label
            with Horizontal(classes="option_row"):
                yield NoFocusLabel("Keymap Name:")
                with Vertical():
                    yield self.name_input
                    yield self.name_validation_label
            with Horizontal(id="button_row"):
                yield Button(label="Keep Editing", id="cancel")
                yield Button(label="Discard + Quit", variant="error", id="discard")
                yield self.save_button
        if validation_result.is_valid:
            self.save_button.focus()
        else:
            self.name_input.focus()

    @on(Input.Changed, "#path_input")
    def validate_path(self, event: Input.Changed) -> None:
        if event.validation_result:
            if event.validation_result.is_valid:
                self.path_validation_label.update("")
                self.path_input_valid = True
            else:
                self.path_validation_label.update(
                    " ".join(event.validation_result.failure_descriptions)
                )
                self.path_input_valid = False

    @on(Input.Changed, "#name_input")
    def validate_name(self, event: Input.Changed) -> None:
        if event.validation_result:
            if event.validation_result.is_valid:
                self.name_validation_label.update("")
                self.name_input_valid = True
            else:
                self.name_validation_label.update(
                    " ".join(event.validation_result.failure_descriptions)
                )
                self.name_input_valid = False

    @on(Input.Submitted)
    def save_from_input(self) -> None:
        self.action_save()

    @on(Button.Pressed, "#submit")
    def save_from_button(self) -> None:
        self.action_save()

    @on(Button.Pressed, "#discard")
    def discard(self) -> None:
        self.dismiss((True, None, None))

    @on(Button.Pressed, "#cancel")
    def cancel(self) -> None:
        self.action_cancel()

    def action_save(self) -> None:
        if self.path_input_valid and self.name_input_valid:
            self.dismiss((True, Path(self.path_input.value), self.name_input.value))

    def action_cancel(self) -> None:
        self.dismiss((False, None, None))


class InputModal(ModalScreen[str], inherit_bindings=False):
    def __init__(self, source_button: Widget) -> None:
        super().__init__()
        self.source_button = source_button

    def compose(self) -> ComposeResult:
        outer = Vertical(id="outer")
        with outer:
            yield NoFocusLabel("Press a key combination")

    @on(Key)
    def close_and_post_key(self, event: Key) -> None:
        event.stop()
        event.prevent_default()
        self.dismiss(result=event.key)


class EditModal(ModalScreen):
    BINDINGS = [
        Binding("escape", "cancel", "Cancel"),
        Binding("up,left", "focus_previous", "focus_previous", show=False),
        Binding("down,right", "focus_next", "focus_next", show=False),
    ]

    def __init__(self, binding: HarlequinKeyBinding) -> None:
        super().__init__()
        self.binding = binding

    def compose(self) -> ComposeResult:
        outer = Vertical(id="outer")
        with outer:
            yield Static(
                (
                    f"[b]Action: {format_action(self.binding.action)}[/b]\n\n"
                    "[dim][i]Use the buttons below to replace, remove, or add "
                    "bindings for this action.[/i]\n"
                    "[b]Enter[/b]: Select Button; [b]Tab[/b]: Next Button[/dim]"
                ),
                id="instructions",
            )
            for key in [key for key in self.binding.keys.split(",") if key]:
                with Horizontal(classes="option_row"):
                    yield NoFocusLabel("Key:")
                    yield EditButton(key=key).focus()
                    yield RemoveButton(key=key)
            with Horizontal(classes="option_row", id="add_row"):
                yield NoFocusLabel("")
                yield AddButton(label="Add Key")
            with Horizontal(classes="option_row"):
                yield NoFocusLabel("Key Display:")
                yield Input(
                    value=(
                        self.binding.key_display if self.binding.key_display else None
                    ),
                    placeholder="(Optional)",
                    id="key_display_input",
                )
            with Horizontal(id="button_row"):
                yield Button(label="Cancel", variant="error", id="cancel")
                yield Button(label="Submit", variant="primary", id="submit")

    def action_cancel(self) -> None:
        self.dismiss(result=self.binding)

    def action_submit(self) -> None:
        keys = ",".join(sorted({btn.key for btn in self.query(EditButton)}))
        key_display: str | None = self.query_one(
            "#key_display_input", expect_type=Input
        ).value
        if not key_display:
            key_display = None
        new_binding = HarlequinKeyBinding(
            keys=keys, action=self.binding.action, key_display=key_display
        )
        self.dismiss(result=new_binding)

    @on(Button.Pressed, "#cancel")
    def cancel_modal(self) -> None:
        self.action_cancel()

    @on(Button.Pressed, "#submit")
    def submit(self) -> None:
        self.action_submit()

    @on(Button.Pressed)
    def handle_button_press(self, message: Button.Pressed) -> None:
        if isinstance(message.button, EditButton):

            def edit_button(new_key: str | None) -> None:
                if new_key is None:
                    return
                assert isinstance(message.button, EditButton)
                message.button.label = new_key
                message.button.key = new_key

            self.app.push_screen(InputModal(source_button=message.button), edit_button)
            return

        elif isinstance(message.button, AddButton):

            def add_button(key: str | None) -> None:
                if key is None:
                    return
                row = Horizontal(
                    NoFocusLabel("Key:"),
                    EditButton(key=key),
                    RemoveButton(key=key),
                    classes="option_row",
                )
                self.mount(row, before="#add_row")

            self.app.push_screen(InputModal(source_button=message.button), add_button)
            return

        elif isinstance(message.button, RemoveButton):
            row = message.button.parent
            assert isinstance(row, Horizontal)
            row.remove()
            return


class HarlequinKeys(AppBase):
    BINDINGS = [Binding("ctrl+q", "quit", "Quit")]
    CSS_PATH = ["global.tcss", "keys_app.tcss"]

    def __init__(
        self,
        *,
        theme: str | None = "harlequin",
        config_path: Path | None = None,
        profile_name: str | None = None,
        keymap_name: list[str] | None = None,
        driver_class: type[Driver] | None = None,
        watch_css: bool = False,
    ):
        super().__init__(
            theme=theme,
            driver_class=driver_class,
            watch_css=watch_css,
        )
        self.config_path = config_path
        self.profile_name = profile_name
        self.plugin_keymap_names: list[str] | None = None
        self.active_keymap_names = keymap_name if keymap_name else None
        self.bindings: dict[str, HarlequinKeyBinding] | None = None
        self.unmodifed_bindings: dict[str, HarlequinKeyBinding] | None = None
        self.table: DataTable | None = None

    def on_mount(self) -> None:
        self.load_bindings()

    def compose(self) -> ComposeResult:
        self.loading_indicator: LoadingIndicator | None = LoadingIndicator()
        yield self.loading_indicator
        # yield self.search_input
        yield Footer(show_command_palette=False)

    def push_edit_modal(self, binding: HarlequinKeyBinding, cursor_row: int) -> None:
        def update_binding(new_binding: HarlequinKeyBinding | None) -> None:
            if new_binding is None:
                return
            assert self.bindings is not None
            assert self.table is not None
            k = format_action(new_binding.action)
            existing_binding = self.bindings[k]
            self.bindings[k] = new_binding
            if existing_binding.keys != new_binding.keys:
                self.table.update_cell(
                    row_index=cursor_row,
                    column_index=1,
                    value=new_binding.keys,
                    update_width=True,
                )
            if existing_binding.key_display != new_binding.key_display:
                self.table.update_cell(
                    row_index=cursor_row,
                    column_index=2,
                    value=new_binding.key_display or "",
                    update_width=True,
                )

        if len(self.screen_stack) == 1:
            self.push_screen(EditModal(binding=binding), update_binding)

    @on(PluginKeymapsFound)
    def register_plugin_keymap_names(self, message: PluginKeymapsFound) -> None:
        self.plugin_keymap_names = message.keymap_names

    @on(ActiveKeymapFound)
    def update_keymap_name(self, message: ActiveKeymapFound) -> None:
        self.active_keymap_names = message.keymap_names

    @on(BindingsReady)
    def mount_bindings_table(self, message: BindingsReady) -> None:
        self.unmodifed_bindings = message.bindings
        self.bindings = self.unmodifed_bindings.copy()
        self.instructions = Static(
            "[i]The table below shows the active key bindings loaded from your "
            "config. Edit the key bindings, then quit to save them to a keymap.[/i]\n"
            f"[b]Loaded Keymaps: {', '.join(self.active_keymap_names or [])}[/b]",
            id="instructions",
        )
        self.table = BindingTable(
            column_labels=["Action", "Keys", "Key Display"], data=message.table_data
        )
        if self.loading_indicator is not None:
            self.loading_indicator.remove()
            self.loading_indicator = None
        self.mount(self.instructions)
        self.mount(self.table)
        self.table.focus()

    @on(DataTable.RowSelected)
    def show_edit_modal(self, message: DataTable.RowSelected) -> None:
        if self.table is None or self.bindings is None:
            return
        action_name = self.table.get_cell_at(Coordinate(message.cursor_row, 0))
        binding = self.bindings[action_name]
        self.push_edit_modal(binding=binding, cursor_row=message.cursor_row)

    @work(
        thread=True,
        exclusive=True,
        exit_on_error=True,
        group="binding_loaders",
        description="Loading bindings from plug-ins and config",
    )
    def load_bindings(self) -> None:
        displayed_bindings = {
            format_action(action): HarlequinKeyBinding(
                keys="", action=action, key_display=""
            )
            for action in HARLEQUIN_ACTIONS
        }

        try:
            profile, user_keymaps = get_config_for_profile(
                config_path=self.config_path, profile_name=self.profile_name
            )
        except HarlequinError as e:
            self.exit(return_code=2, message=pretty_error_message(e))
        all_keymaps = load_keymap_plugins(user_defined_keymaps=user_keymaps)
        plugin_keymap_names = [
            keymap.name
            for keymap in all_keymaps.values()
            if keymap.name not in [k.name for k in user_keymaps]
        ]
        self.post_message(PluginKeymapsFound(names=plugin_keymap_names))
        profile_keymap_names = profile.get("keymap_name")
        active_keymap_names = (
            self.active_keymap_names or profile_keymap_names or ["vscode"]
        )
        if self.active_keymap_names is None:
            self.post_message(ActiveKeymapFound(names=active_keymap_names))
        for keymap_name in active_keymap_names:
            keymap = all_keymaps.get(keymap_name)
            if keymap is None:
                continue
            for binding in keymap.bindings:
                merged_action = displayed_bindings[format_action(binding.action)]
                deduped_keys = (
                    set(merged_action.keys.split(","))
                    .union(set(binding.keys.split(",")))
                    .difference(set([""]))
                )
                merged_action.keys = ",".join(sorted(deduped_keys))
                if binding.key_display:
                    merged_action.key_display = binding.key_display

        table_data = format_bindings_for_table(displayed_bindings=displayed_bindings)
        self.post_message(
            BindingsReady(bindings=displayed_bindings, table_data=table_data)
        )

    async def action_quit(self) -> None:
        if (
            (not self.active_keymap_names)
            or self.bindings is None
            or self.unmodifed_bindings is None
            or self.plugin_keymap_names is None
        ):
            # app hasn't finished loading yet, just quit!
            await super().action_quit()
            return  # for mypy
        modified_bindings = [
            new_binding
            for new_binding, old_binding in zip(
                self.bindings.values(), self.unmodifed_bindings.values()
            )
            if new_binding != old_binding
        ]
        if not modified_bindings:
            await super().action_quit()
            return  # for mypy

        def maybe_save(
            screen_data: tuple[bool, Path | None, str | None] | None,
        ) -> None:
            if screen_data is None:
                return
            do_quit, config_path, keymap_name = screen_data
            if not do_quit:
                return
            if not config_path or not keymap_name:
                # the user pressed "Discard" instead of "Save"
                self.app.exit()
                return  # for mypy
            config_file = ConfigFile(path=config_path)
            keymaps = config_file.relevant_config.get("keymaps", {})
            assert self.bindings is not None
            keymaps.update(
                {
                    keymap_name: [
                        b.to_dict()
                        for b in self.bindings.values()
                        if b.keys or b.key_display
                    ]
                }
            )
            config_file.update({"keymaps": keymaps})
            config_file.write()
            self.app.exit(
                message=Panel.fit(
                    (
                        f"Keymap {keymap_name} written to file at {config_path}\n"
                        "To use this keymap, invoke harlequin with this option:\n"
                        f"harlequin --keymap-name {keymap_name}\n"
                        "Or add this line to a profile in your TOML config file:\n"
                        f"keymap_name=['{keymap_name}']"
                    ),
                    title="Keymap successfully created",
                    title_align="left",
                    border_style=YELLOW,
                )
            )

        self.push_screen(
            QuitModal(
                config_path=self.config_path,
                keymap_name=self.active_keymap_names[-1],
                plugin_names=self.plugin_keymap_names,
            ),
            maybe_save,
        )


def format_action(action: str) -> str:
    component_name, _, action_name = action.rpartition(".")
    component_display_name = (
        f"{' '.join(w.capitalize() for w in component_name.split('_'))}: "
        if component_name
        else ""
    )
    action_display_name = " ".join(w.capitalize() for w in action_name.split("_"))
    return f"{component_display_name}{action_display_name}"


def format_bindings_for_table(
    displayed_bindings: dict[str, HarlequinKeyBinding],
) -> list[tuple[str, str, str]]:
    table_data: list[tuple[str, str, str]] = []
    for formatted_name, binding in displayed_bindings.items():
        table_data.append(
            (
                formatted_name,
                binding.keys,
                binding.key_display or "",
            )
        )
    return table_data



from __future__ import annotations

import os

from questionary import Style as QuestionaryStyle
from textual.theme import BUILTIN_THEMES
from textual.theme import Theme as TextualTheme

GREEN = "#45FFCA"
YELLOW = "#FEFFAC"
PINK = "#FFB6D9"
PURPLE = "#D67BFF"
GRAY = "#777777"
DARK_GRAY = "#555555"
BLACK = "#0C0C0C"
WHITE = "#DDDDDD"

HARLEQUIN_TEXTUAL_THEME = TextualTheme(
    name="harlequin",
    primary=YELLOW,
    secondary=GREEN,
    warning=YELLOW,
    error=PINK,
    success=GREEN,
    accent=PINK,
    foreground=WHITE,
    background=BLACK,
    surface=BLACK,
    panel=DARK_GRAY,
    dark=True,
)

VALID_THEMES = BUILTIN_THEMES
VALID_THEMES.pop("textual-ansi")
VALID_THEMES.update({"harlequin": HARLEQUIN_TEXTUAL_THEME})

HARLEQUIN_QUESTIONARY_STYLE = (
    QuestionaryStyle(
        [
            ("qmark", "fg:ansidefault bold"),
            ("question", "fg:ansidefault nobold"),
            ("answer", "fg:ansidefault bold"),
            ("pointer", "fg:ansidefault bold"),
            ("highlighted", "fg:ansidefault bold"),
            ("selected", "fg:ansidefault noreverse bold"),
            ("separator", "fg:ansidefault"),
            ("instruction", "fg:ansidefault italic"),
            ("text", ""),
            ("disabled", "fg:ansidefault italic"),
        ]
    )
    if os.getenv("NO_COLOR")
    else QuestionaryStyle(
        [
            ("qmark", f"fg:{GREEN} bold"),
            ("question", "bold"),
            ("answer", f"fg:{YELLOW} bold"),
            ("pointer", f"fg:{YELLOW} bold"),
            ("highlighted", f"fg:{YELLOW} bold"),
            ("selected", f"fg:{YELLOW} noreverse bold"),
            ("separator", f"fg:{PURPLE}"),
            ("instruction", "fg:#858585 italic"),
            ("text", ""),
            ("disabled", "fg:#858585 italic"),
        ]
    )
)



from __future__ import annotations

from typing import Type, Union

from textual.app import App, InvalidThemeError
from textual.binding import ActiveBinding
from textual.driver import Driver
from textual.screen import Screen
from textual.types import CSSPathType

from harlequin.colors import HARLEQUIN_TEXTUAL_THEME
from harlequin.exception import (
    HarlequinThemeError,
    pretty_error_message,
)


class ScreenBase(Screen):
    @property
    def active_bindings(self) -> dict[str, ActiveBinding]:
        def sort_key(binding_pair: tuple[str, ActiveBinding]) -> int:
            return 0 if binding_pair[1].node == self.app else 1

        binding_map = {
            k: v for k, v in sorted(super().active_bindings.items(), key=sort_key)
        }
        return binding_map


class AppBase(App, inherit_bindings=False):
    """
    A common base app for Harlequin and its mini-apps.
    """

    def __init__(
        self,
        *,
        theme: str | None = "harlequin",
        driver_class: Union[Type[Driver], None] = None,
        css_path: Union[CSSPathType, None] = None,
        watch_css: bool = False,
    ):
        super().__init__(driver_class, css_path, watch_css)
        self.register_theme(HARLEQUIN_TEXTUAL_THEME)
        try:
            self.theme = theme or "harlequin"
        except InvalidThemeError:
            from harlequin.colors import VALID_THEMES

            valid_themes = ", ".join(VALID_THEMES.keys())
            e = HarlequinThemeError(
                (
                    f"No theme found with the name {theme}.\n"
                    "Supported themes changed in Harlequin v2.0.0. "
                    "Theme must be `harlequin` or the name of a Textual Theme:\n"
                    f"{valid_themes}"
                ),
                title="Harlequin couldn't load your theme.",
            )
            self.exit(return_code=2, message=pretty_error_message(e))

    def get_default_screen(self) -> Screen:
        """
        Changes the default screen to re-order bindings, with global bindings first.
        """
        return ScreenBase(id="_default")

    def _handle_exception(self, error: Exception) -> None:
        """
        Prevents tracebacks from being printed due to exceptions that
        occur after App.exit() is called.
        See https://github.com/Textualize/textual/issues/5325
        """
        if self._exit:
            return
        return super()._handle_exception(error)



from __future__ import annotations

import pytest

from harlequin.components.data_catalog.s3_tree import S3Tree


@pytest.mark.parametrize(
    "uri,expected",
    [
        ("my-bucket", (None, "my-bucket", "")),
        ("my-bucket/my-prefix", (None, "my-bucket", "my-prefix")),
        ("s3://my-bucket", (None, "my-bucket", "")),
        ("s3://my-bucket/my-prefix", (None, "my-bucket", "my-prefix")),
        (
            "https://my-storage.com/my-bucket/",
            ("https://my-storage.com", "my-bucket", ""),
        ),
        (
            "https://my-storage.com/my-bucket/my-prefix",
            ("https://my-storage.com", "my-bucket", "my-prefix"),
        ),
        (
            "https://storage.my-api.com/my-bucket/",
            ("https://storage.my-api.com", "my-bucket", ""),
        ),
        ("https://my-bucket.s3.amazonaws.com", (None, "my-bucket", "")),
        (
            "https://my-bucket.s3.amazonaws.com/my-prefix",
            (None, "my-bucket", "my-prefix"),
        ),
        (
            "https://my-bucket.storage.googleapis.com",
            ("https://storage.googleapis.com", "my-bucket", ""),
        ),
        (
            "https://my-bucket.storage.googleapis.com/my-prefix",
            ("https://storage.googleapis.com", "my-bucket", "my-prefix"),
        ),
    ],
)
def test_s3_uri_parsing(uri: str, expected: tuple[str | None, str, str]) -> None:
    actual = S3Tree._parse_s3_uri(uri)
    assert actual == expected



from __future__ import annotations

import sys
from datetime import date, datetime
from typing import Awaitable, Callable
from unittest.mock import MagicMock

import pytest
from textual.message import Message
from textual_fastdatatable import DataTable

from harlequin import Harlequin
from harlequin.components.results_viewer import ResultsViewer


def transaction_button_visible(app: Harlequin) -> bool:
    """
    Skip snapshot checks for versions of that app showing the autocommit button.
    """
    return sys.version_info >= (3, 12) and "Sqlite" in app.adapter.__class__.__name__


@pytest.mark.asyncio
async def test_dupe_column_names(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    query = "select 1 as a, 1 as a, 2 as a, 2 as a"
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        if not transaction_button_visible(app):
            assert await app_snapshot(app, "dupe columns")


@pytest.mark.asyncio
async def test_copy_data(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
    mock_pyperclip: MagicMock,
) -> None:
    app = app_all_adapters
    query = "select 3, 'rosberg', 6, 'ROS', 'Nico', 'Rosberg', '1985-06-27', 'German', 'http://en.wikipedia.org/wiki/Nico_Rosberg'"
    expected = "3	rosberg	6	ROS	Nico	Rosberg	1985-06-27	German	http://en.wikipedia.org/wiki/Nico_Rosberg"
    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()

        assert app.results_viewer._has_focus_within
        keys = ["shift+right"] * 8
        await pilot.press(*keys)
        await pilot.wait_for_scheduled_animations()
        await pilot.press("ctrl+c")
        await pilot.pause()

        copied_message = list(
            filter(lambda m: isinstance(m, DataTable.SelectionCopied), messages)
        )[0]
        assert isinstance(copied_message, DataTable.SelectionCopied)
        assert isinstance(copied_message.values, list)

        app.editor.text = ""
        app.editor.focus()
        await pilot.press("ctrl+v")  # paste
        assert app.editor.text == expected
        if not transaction_button_visible(app):
            assert await app_snapshot(app, "paste values from table")


@pytest.mark.asyncio
async def test_data_truncated_with_tooltip(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    query = "select 'supercalifragilisticexpialidocious'"
    async with app.run_test(tooltips=True) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()

        await pilot.hover(ResultsViewer, (2, 2))
        await pilot.pause(0.5)
        if not transaction_button_visible(app):
            assert await app_snapshot(app, "hover over truncated value")


@pytest.mark.asyncio
async def test_infinity_timestamp(
    app: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    query = """
        select
            'infinity'::date,
            'infinity'::timestamp,
            '-infinity'::date,
            '-infinity'::timestamp
        """
    async with app.run_test(size=(120, 36)) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()

        results_table = app.results_viewer.get_visible_table()
        assert results_table is not None
        assert results_table.get_row_at(0) == [
            date.max,
            datetime.max,
            date.min,
            datetime.min,
        ]

        assert await app_snapshot(app, "hover over truncated value")



import pickle
from pathlib import Path
from typing import List

import pytest
from textual.widgets.text_area import Selection

from harlequin import Harlequin
from harlequin.editor_cache import (
    BufferState,
    Cache,
    get_cache_file,
    load_cache,
    write_cache,
)


@pytest.fixture
def buffer_states() -> List[BufferState]:
    return [
        BufferState(
            selection=Selection((0, 3), (0, 3)),
            text="select 1\n",
        ),
        BufferState(
            selection=Selection((0, 0), (0, 0)),
            text="",
        ),
        BufferState(
            selection=Selection((0, 0), (1, 0)),
            text="select\n*\nfrom\nfoo\n",
        ),
    ]


@pytest.fixture
def cache(buffer_states: List[BufferState]) -> Cache:
    return Cache(focus_index=1, buffers=buffer_states)


@pytest.fixture(autouse=True)
def mock_user_cache_dir(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Path:
    monkeypatch.setattr("harlequin.editor_cache.user_cache_dir", lambda **_: tmp_path)
    return tmp_path


@pytest.mark.use_cache
def test_cache_ops(mock_user_cache_dir: Path, cache: Cache) -> None:
    assert mock_user_cache_dir.exists()
    assert len(list(mock_user_cache_dir.iterdir())) == 0
    write_cache(cache)
    children = list(mock_user_cache_dir.iterdir())
    assert len(children) == 1
    assert children[0].suffix == ".pickle"
    assert get_cache_file() == children[0]
    loaded_cache = load_cache()
    assert loaded_cache == cache


@pytest.mark.use_cache
@pytest.mark.asyncio
async def test_harlequin_loads_cache(cache: Cache, app: Harlequin) -> None:
    write_cache(cache)
    async with app.run_test() as pilot:
        while app.editor is None:
            await pilot.pause()
        assert app.editor_collection is not None
        assert app.editor is not None
        assert app.editor_collection.tab_count == len(cache.buffers)
        assert [editor.text for editor in app.editor_collection.all_editors] == [
            buffer.text for buffer in cache.buffers
        ]


@pytest.mark.use_cache
@pytest.mark.asyncio
async def test_harlequin_writes_cache(app: Harlequin) -> None:
    cache_path = get_cache_file()
    assert not cache_path.exists()
    async with app.run_test() as pilot:
        while app.editor is None:
            await pilot.pause()
        assert app.editor_collection is not None
        assert app.editor_collection.tab_count == 1
        app.editor.text = "first"
        await pilot.press("ctrl+n")
        await pilot.pause()
        app.editor.text = "second"
        await pilot.press("ctrl+q")
    assert cache_path.exists()
    with open(cache_path, "rb") as f:
        cache = pickle.load(f)
    assert isinstance(cache, Cache)
    assert [buffer.text for buffer in cache.buffers] == ["first", "second"]



import asyncio
from unittest.mock import patch

from textual.widgets.text_area import Selection

from harlequin import Harlequin
from harlequin.editor_cache import BufferState, Cache
from harlequin_duckdb import DuckDbAdapter


async def load_lots_of_buffers() -> None:
    with patch("harlequin.components.code_editor.load_cache") as mock_load_cache:
        buff = BufferState(
            selection=Selection((0, 0), (0, 0)),
            text="select 1000; ",
        )
        cache = Cache(focus_index=0, buffers=[buff])
        mock_load_cache.return_value = cache

        adapter = DuckDbAdapter((":memory:",), no_init=True)
        app = Harlequin(adapter=adapter)

        async with app.run_test() as pilot:
            while app.editor is None:
                await pilot.pause()
            await pilot.press("ctrl+j")
            while (table := app.results_viewer.get_visible_table()) is None:
                await pilot.pause()
            assert table.row_count == 1

        app.exit()


asyncio.run(load_lots_of_buffers())
print("Ran successfully")



from __future__ import annotations

from textwrap import dedent
from typing import Awaitable, Callable
from unittest.mock import MagicMock

import pytest
from textual.css.query import NoMatches

from harlequin import Harlequin

QUERY = dedent(
    """
    select *
    from
        (
            values
                (1, 2, 3),
                (4, 5, 6),
                (7, 8, 9),
                (10, 11, 12),
                (13, 14, 15),
                (16, 17, 18),
                (19, 20, 21)
        ) foo(a, b, c)
"""
).strip()


@pytest.mark.asyncio
async def test_editor_bindings(
    app: Harlequin,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
    mock_pyperclip: MagicMock,
) -> None:
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()

        q = QUERY
        app.editor.text = q
        assert app.editor.selection.start == app.editor.selection.end == (0, 0)

        # simple navigation
        await pilot.press("down")
        assert app.editor.selection.start == app.editor.selection.end == (1, 0)
        await pilot.press("right")
        assert app.editor.selection.start == app.editor.selection.end == (1, 1)
        await pilot.press("right")
        assert app.editor.selection.start == app.editor.selection.end == (1, 2)
        await pilot.press("left")
        assert app.editor.selection.start == app.editor.selection.end == (1, 1)
        await pilot.press("up")
        assert app.editor.selection.start == app.editor.selection.end == (0, 1)
        await pilot.press("ctrl+right")
        assert app.editor.selection.start == app.editor.selection.end == (0, 6)
        await pilot.press("ctrl+left")
        assert app.editor.selection.start == app.editor.selection.end == (0, 0)
        await pilot.press("ctrl+end")
        assert app.editor.selection.start == app.editor.selection.end == (11, 18)
        await pilot.press("ctrl+home")
        assert app.editor.selection.start == app.editor.selection.end == (0, 0)

        # simple selection
        await pilot.press("shift+down")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (1, 0)
        await pilot.press("shift+right")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (1, 1)
        await pilot.press("shift+right")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (1, 2)
        await pilot.press("shift+left")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (1, 1)
        await pilot.press("shift+up")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (0, 1)
        await pilot.press("ctrl+shift+right")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (0, 6)
        await pilot.press("ctrl+shift+left")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (0, 0)
        await pilot.press("ctrl+shift+end")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (11, 18)
        await pilot.press("ctrl+shift+home")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (0, 0)
        await pilot.press("ctrl+a")
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (11, 18)

        # cut/copy/paste
        await pilot.press("ctrl+c")
        assert app.editor.text == QUERY
        assert app.editor.text_input is not None
        assert app.editor.text_input.clipboard == QUERY
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (11, 18)
        await pilot.press("ctrl+x")
        assert app.editor.text == ""
        assert app.editor.text_input.clipboard == QUERY
        assert app.editor.selection.start == (0, 0)
        assert app.editor.selection.end == (0, 0)
        await pilot.press("ctrl+v")
        assert app.editor.text == QUERY
        assert app.editor.selection.start == app.editor.selection.end == (11, 18)

        await pilot.press("a")
        assert app.editor.text == QUERY + "a"
        await pilot.press("escape")  # dismiss autocomplete
        await pilot.press("enter")
        assert app.editor.text == QUERY + "a\n    "

        # undo/redo
        await pilot.press("ctrl+z")
        assert app.editor.text == QUERY + "a"
        await pilot.press("ctrl+y")
        assert app.editor.text == QUERY + "a\n    "

        # delete
        await pilot.press("backspace")
        assert app.editor.text == QUERY + "a\n   "
        await pilot.press("shift+delete")
        assert app.editor.text == QUERY + "a"
        await pilot.press("backspace")
        assert app.editor.text == QUERY
        await pilot.press("ctrl+home")
        await pilot.press("delete")
        assert app.editor.text == QUERY[1:]

        # find
        await pilot.press("ctrl+f")
        assert app.query_one("#textarea__find_input")
        await pilot.press("escape")
        with pytest.raises(NoMatches):
            _ = app.query_one("#textarea__find_input")
        await pilot.press("f3")
        assert app.query_one("#textarea__find_input")
        await pilot.press("escape")
        with pytest.raises(NoMatches):
            _ = app.query_one("#textarea__find_input")

        # goto line
        await pilot.press("ctrl+g")
        assert app.query_one("#textarea__gotoline_input")
        await pilot.press("escape")
        with pytest.raises(NoMatches):
            _ = app.query_one("#textarea__gotoline_input")

        # save
        await pilot.press("ctrl+s")
        assert app.query_one("#textarea__save_input")
        await pilot.press("escape")
        with pytest.raises(NoMatches):
            _ = app.query_one("#textarea__save_input")

        # open
        await pilot.press("ctrl+o")
        assert app.query_one("#textarea__open_input")
        await pilot.press("escape")
        with pytest.raises(NoMatches):
            _ = app.query_one("#textarea__open_input")


@pytest.mark.asyncio
async def test_results_viewer_bindings(
    app: Harlequin,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
    mock_pyperclip: MagicMock,
) -> None:
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()

        q = QUERY
        app.editor.text = q
        await pilot.press("ctrl+j")

        while (table := app.results_viewer.get_visible_table()) is None:
            await pilot.pause()

        assert table is not None
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate is None

        # simple navigation
        await pilot.press("down")
        assert table.cursor_coordinate == (1, 0)
        assert table.selection_anchor_coordinate is None
        await pilot.press("right")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate is None
        await pilot.press("right")
        assert table.cursor_coordinate == (1, 2)
        assert table.selection_anchor_coordinate is None
        await pilot.press("left")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate is None
        await pilot.press("up")
        assert table.cursor_coordinate == (0, 1)
        assert table.selection_anchor_coordinate is None
        await pilot.press("ctrl+right")
        assert table.cursor_coordinate == (0, 2)
        assert table.selection_anchor_coordinate is None
        await pilot.press("ctrl+left")
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate is None
        await pilot.press("ctrl+end")
        assert table.cursor_coordinate == (6, 2)
        assert table.selection_anchor_coordinate is None
        await pilot.press("ctrl+home")
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate is None

        # simple selection
        await pilot.press("shift+down")
        assert table.cursor_coordinate == (1, 0)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("shift+right")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("shift+right")
        assert table.cursor_coordinate == (1, 2)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("shift+left")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("shift+up")
        assert table.cursor_coordinate == (0, 1)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("ctrl+shift+right")
        assert table.cursor_coordinate == (0, 2)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("ctrl+shift+left")
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("ctrl+shift+end")
        assert table.cursor_coordinate == (6, 2)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("ctrl+shift+home")
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate == (0, 0)
        await pilot.press("ctrl+a")
        assert table.cursor_coordinate == (6, 2)
        assert table.selection_anchor_coordinate == (0, 0)

        # copy
        await pilot.press("ctrl+c")
        assert app.editor.text_input is not None
        assert app.editor.text_input.clipboard.startswith("1\t2\t3")



from __future__ import annotations

import sqlite3
from contextlib import suppress
from itertools import cycle, zip_longest
from pathlib import Path
from typing import Any, Literal, Sequence
from urllib.parse import unquote, urlparse

from textual_fastdatatable.backend import AutoBackendType

from harlequin.adapter import HarlequinAdapter, HarlequinConnection, HarlequinCursor
from harlequin.autocomplete.completion import HarlequinCompletion
from harlequin.catalog import Catalog, CatalogItem
from harlequin.exception import (
    HarlequinConfigError,
    HarlequinConnectionError,
    HarlequinQueryError,
)
from harlequin.options import HarlequinAdapterOption, HarlequinCopyFormat
from harlequin.transaction_mode import HarlequinTransactionMode
from harlequin_sqlite.catalog import DatabaseCatalogItem
from harlequin_sqlite.cli_options import SQLITE_OPTIONS
from harlequin_sqlite.completions import get_completion_data

IN_MEMORY_CONN_STR = (":memory:",)


class HarlequinSqliteCursor(HarlequinCursor):
    def __init__(self, conn: HarlequinSqliteConnection, cur: sqlite3.Cursor) -> None:
        self.conn = conn
        self.cur = cur
        self._limit: int | None = None
        try:
            _first_row = cur.fetchone()
        except sqlite3.Error:  # maybe canceled query here
            _first_row = None
        self.has_records = _first_row is not None
        self._first_row: tuple[Any, ...] = _first_row or tuple(
            [None] * len(cur.description)
        )

    def columns(self) -> list[tuple[str, str]]:
        col_names = [col[0] for col in self.cur.description]
        col_types = [
            self.conn._short_column_type_from_python_object(value)
            for value in self._first_row
        ]
        return list(zip_longest(col_names, col_types, fillvalue="?"))

    def set_limit(self, limit: int) -> "HarlequinSqliteCursor":
        self._limit = limit
        return self

    def fetchall(self) -> AutoBackendType | None:
        if self.has_records:
            try:
                remaining_rows = (
                    self.cur.fetchall()
                    if self._limit is None
                    else self.cur.fetchmany(self._limit - 1)
                )
            except sqlite3.OperationalError:  # maybe canceled here
                return None
            except sqlite3.Error as e:
                raise HarlequinQueryError(
                    msg=str(e),
                    title=(
                        "SQLite raised an error when fetching results for your query:"
                    ),
                ) from e
            return [self._first_row, *remaining_rows]
        else:
            return None

    def fetchone(self) -> tuple | None:
        return self._first_row


class HarlequinSqliteConnection(HarlequinConnection):
    def __init__(self, conn: sqlite3.Connection, init_message: str = "") -> None:
        self.conn = conn
        self.init_message = init_message
        self._transaction_modes: list[HarlequinTransactionMode | None] = (
            [
                HarlequinTransactionMode(label="Auto"),
                HarlequinTransactionMode(
                    label="Manual", commit=self.conn.commit, rollback=self.conn.rollback
                ),
            ]
            if hasattr(conn, "autocommit")
            else [None]
        )
        self._transaction_mode_gen = cycle(self._transaction_modes)
        self._transaction_mode = next(self._transaction_mode_gen)
        self._sync_connection_transaction_mode()

    def execute(self, query: str) -> HarlequinSqliteCursor | None:
        # the behavior on manual mode is really counter-intuitive; if a
        # transaction isn't explicitly began, it's basically the same as
        # auto. By forcing an explicit begin, the behavior is more like
        # manual mode on other databases.
        if (
            self.transaction_mode
            and self.transaction_mode.label == "Manual"
            and not self.conn.in_transaction
        ):
            with suppress(sqlite3.Error):
                self.conn.execute("begin;")
        try:
            cur = self.conn.execute(query)
        except sqlite3.Error as e:
            # no-op if user ran a begin; statement, since we already started the txn
            if (
                isinstance(e, sqlite3.OperationalError)
                and str(e) == "cannot start a transaction within a transaction"
            ):
                return None
            raise HarlequinQueryError(
                msg=str(e),
                title="SQLite raised an error when compiling or running your query:",
            ) from e

        if cur.description is not None:
            return HarlequinSqliteCursor(conn=self, cur=cur)
        else:
            return None

    def cancel(self) -> None:
        self.conn.interrupt()

    def get_catalog(self) -> Catalog:
        catalog_items: list[CatalogItem] = []
        databases = self._get_databases()
        for database_label in databases:
            catalog_items.append(
                DatabaseCatalogItem.from_label(label=database_label, connection=self)
            )
        return Catalog(items=catalog_items)

    def get_completions(self) -> list[HarlequinCompletion]:
        return get_completion_data(self.conn)

    @property
    def transaction_mode(self) -> HarlequinTransactionMode | None:
        return self._transaction_mode

    def toggle_transaction_mode(self) -> HarlequinTransactionMode | None:
        new_mode = next(self._transaction_mode_gen)
        self._transaction_mode = new_mode
        self._sync_connection_transaction_mode()
        return new_mode

    def close(self) -> None:
        self.conn.close()

    def _sync_connection_transaction_mode(self) -> None:
        if not self._transaction_mode or not hasattr(self.conn, "autocommit"):
            return

        if self.conn.in_transaction:
            self.conn.commit()

        if self.transaction_mode and self.transaction_mode.label == "Auto":
            self.conn.autocommit = True
        else:
            self.conn.autocommit = False

    def _get_databases(self) -> list[str]:
        objects: list[tuple[str, str, str]] = self.conn.execute(
            "pragma database_list"
        ).fetchall()
        return [db_name for _, db_name, _ in objects]

    def _get_relations(self, db_name: str) -> list[tuple[str, str]]:
        objects = self.conn.execute(
            f'select type, name from "{db_name}".sqlite_schema'
        ).fetchall()
        relations = [(name, typ) for typ, name in objects if typ in ("table", "view")]
        return relations

    def _get_columns(self, db_name: str, rel_name: str) -> list[tuple[str, str, str]]:
        return self.conn.execute(
            f"pragma {db_name}.table_info('{rel_name}')"
        ).fetchall()

    @staticmethod
    def _short_column_type(raw_type: str) -> str:
        # first get the storage affinity for a type decl
        if raw_type is None or raw_type == "":
            affinity = ""

        typ = raw_type.lower()
        if "int" in typ:
            affinity = "INTEGER"
        elif any([x in typ for x in ["char", "clob", "text"]]):
            affinity = "TEXT"
        elif "blob" in typ:
            affinity = "BLOB"
        elif any([x in typ for x in ["real", "floa", "doub"]]):
            affinity = "REAL"
        else:
            affinity = "NUMERIC"

        # then lookup the abbreviation for the affinity
        mapping = {
            "": "",
            "TEXT": "s",
            "NUMERIC": "#.#",
            "INTEGER": "##",
            "REAL": "#.#",
            "BLOB": "b",
        }
        return mapping.get(affinity, "?")

    @staticmethod
    def _short_column_type_from_python_object(obj: object) -> str:
        mapping: dict[type | None, str] = {
            None: "",
            str: "s",
            float: "#.#",
            int: "##",
            object: "b",
        }
        return mapping.get(type(obj), "?")

    @staticmethod
    def _short_relation_type(raw_type: str) -> str:
        mapping = {"table": "t", "view": "v"}
        return mapping.get(raw_type, "?")

    def copy(
        self, query: str, path: Path, format_name: str, options: dict[str, Any]
    ) -> None:
        raise NotImplementedError

    def validate_sql(self, text: str) -> str:
        raise NotImplementedError


class HarlequinSqliteAdapter(HarlequinAdapter):
    ADAPTER_OPTIONS: list[HarlequinAdapterOption] | None = SQLITE_OPTIONS
    COPY_FORMATS: list[HarlequinCopyFormat] | None = None
    IMPLEMENTS_CANCEL = True

    def __init__(
        self,
        conn_str: Sequence[str],
        init_path: Path | str | None = None,
        no_init: bool | str = False,
        read_only: bool = False,
        connection_mode: Literal["ro", "rw", "rwc", "memory"] | None = None,
        timeout: str | float = 5.0,
        detect_types: str | int = 0,
        isolation_level: Literal["DEFERRED", "EXCLUSIVE", "IMMEDIATE"] = "DEFERRED",
        cached_statements: str | int = 128,
        extension: list[str] | None = None,
        **_: Any,
    ) -> None:
        try:
            self.conn_str = (
                conn_str
                if conn_str and conn_str != ("",) and connection_mode != "memory"
                else IN_MEMORY_CONN_STR
            )
            self.init_path = (
                Path(init_path).expanduser().resolve()
                if init_path is not None
                else Path.home() / ".sqliterc"
            )
            self.no_init = bool(no_init)
            self.read_only = bool(read_only)
            self.connection_mode = connection_mode
            self.timeout = float(timeout)
            self.detect_types = int(detect_types)
            self.isolation_level = isolation_level
            self.cached_statements = int(cached_statements)
            self.extensions = extension if extension is not None else []
            self.can_load_extensions = hasattr(
                sqlite3.Connection, "enable_load_extension"
            )
        except (ValueError, TypeError) as e:
            raise HarlequinConfigError(
                msg=f"SQLite adapter received bad config value: {e}",
                title="Harlequin could not initialize the selected adapter.",
            ) from e

        if self.extensions and not self.can_load_extensions:
            raise HarlequinConfigError(
                title="Harlequin could not initialize the selected adapter.",
                msg=(
                    "SQLite adapter received --extension option, but extensions "
                    "are disabled on this SQLite distribution. See "
                    "https://harlequin.sh/docs/sqlite/extensions"
                ),
            )

    @property
    def connection_id(self) -> str | None:
        if self.conn_str == IN_MEMORY_CONN_STR:
            return ""
        return ",".join(
            [Path(conn).resolve().as_posix() for conn in sorted(self.conn_str)]
        )

    def connect(self) -> HarlequinSqliteConnection:
        if (
            self.read_only
            and self.connection_mode is not None
            and self.connection_mode != "ro"
        ):
            raise HarlequinConnectionError(
                "Cannot specify readonly flag and a connection mode."
            )
        elif self.read_only:
            mode_str = "?mode=ro"
        elif self.connection_mode is not None:
            mode_str = f"?mode={self.connection_mode}"
        else:
            mode_str = ""

        # build db strings
        db_uris: list[str] = []
        db_names: list[str] = []
        for s in self.conn_str:
            if s == ":memory:":
                db_uris.append(s)
                db_names.append("memory")
            elif s.startswith("file:"):
                db_uris.append(s)
                db_names.append(Path(unquote(urlparse(s).path)).stem)
            else:
                try:
                    p = Path(s).resolve()
                    db = f"{p.as_uri()}{mode_str}"
                    db_uris.append(db)
                    db_names.append(p.stem)
                except ValueError as e:
                    raise HarlequinConnectionError(
                        f"Cannot build URI from connection string {s}.",
                        title="SQLite couldn't connect to your database.",
                    ) from e
        primary_db, *other_dbs = db_uris
        try:
            conn = sqlite3.connect(
                database=primary_db,
                timeout=self.timeout,
                detect_types=self.detect_types,
                isolation_level=self.isolation_level,
                cached_statements=self.cached_statements,
                check_same_thread=False,
                uri=True,
            )
            # sqlite won't error on connect if you open a file that isn't a sqlite db
            # running pragma database_list will raise, though:
            _ = conn.execute("pragma database_list")
        except sqlite3.Error as e:
            if "file is not a database" in (msg := str(e)):
                msg = (
                    "sqlite raised the following error when trying to open "
                    f"{primary_db}:\n---\n{msg}\n---\n\n"
                    "Did you mean to use Harlequin's DuckDB adapter instead? "
                    f"Maybe try:\nharlequin -a duckdb {' '.join(self.conn_str)}\n"
                    f"or:\nharlequin -P None {' '.join(self.conn_str)}"
                )
            raise HarlequinConnectionError(
                msg=msg,
                title="SQLite couldn't connect to your database.",
            ) from e

        # Python 3.12 added an autocommit attribute, which should be
        # turned on by default
        if hasattr(conn, "autocommit"):
            conn.autocommit = True

        for uri, name in zip(other_dbs, db_names[1:]):
            try:
                conn.execute(f"attach database '{uri}' as {name}")
            except sqlite3.Error as e:
                if "file is not a database" in (msg := str(e)):
                    msg = (
                        "sqlite raised the following error when trying to open "
                        f"{uri}:\n---\n{msg}\n---\n\n"
                        "Did you mean to use Harlequin's DuckDB adapter instead? "
                        f"Maybe try:\nharlequin -a duckdb {' '.join(self.conn_str)}\n"
                        f"or:\nharlequin -P None {' '.join(self.conn_str)}"
                    )
                raise HarlequinConnectionError(
                    msg=msg, title="SQLite couldn't connect to your database."
                ) from e

        if self.can_load_extensions:
            conn.enable_load_extension(True)

        for extension in self.extensions:
            try:
                conn.load_extension(extension)
            except sqlite3.Error as e:
                raise HarlequinConnectionError(
                    str(e), title="SQLite couldn't load your extension."
                ) from e

        init_msg = ""
        if self.init_path is not None and not self.no_init:
            init_script = self._read_init_script(self.init_path)
            count = 0
            for command in self._split_script(init_script):
                rewritten_command = self._rewrite_init_command(command)
                for cmd in rewritten_command.split(";"):
                    if cmd.strip():
                        try:
                            conn.execute(cmd)
                        except sqlite3.Error as e:
                            msg = (
                                f"Attempted to execute script at {self.init_path}. "
                                f"Contents:\n{command}\nRewritten to:\n"
                                f"{rewritten_command}\nCurrently executing:\n"
                                f"{cmd}\nError:\n{e}"
                            )
                            if not self.can_load_extensions and isinstance(
                                e, sqlite3.OperationalError
                            ):
                                msg += (
                                    "\nWarning: Cannot load extensions with this "
                                    "SQLite distribution. See "
                                    "https://harlequin.sh/docs/sqlite/extensions"
                                )
                            raise HarlequinConnectionError(
                                msg,
                                title=(
                                    "SQLite could not execute your initialization "
                                    "script."
                                ),
                            ) from e
                        else:
                            count += 1
            if count > 0:
                init_msg = (
                    f"Executed {count} {'command' if count == 1 else 'commands'} "
                    f"from {self.init_path}"
                )
        return HarlequinSqliteConnection(conn=conn, init_message=init_msg)

    @staticmethod
    def _read_init_script(init_path: Path) -> str:
        try:
            with open(init_path.expanduser(), "r") as f:
                init_script = f.read()
        except OSError:
            init_script = ""
        return init_script

    @staticmethod
    def _split_script(script: str) -> list[str]:
        """
        SQLite init scripts can contain SQL queries or dot commands. The SQL
        queries may contain newlines, but the dot commands are newline-terminated.
        This takes a raw script and returns a list of executable commands
        """
        lines = script.splitlines()
        commands: list[str] = []
        i = 0
        for j, line in enumerate(lines):
            if line.startswith("."):
                commands.append("\n".join(lines[i:j]))
                commands.append(line)
                i = j + 1
            # queries can be terminated by / or "go" on a line
            # by itself
            elif line in ("/", "go"):
                commands.append("\n".join(lines[i:j]))
                i = j + 1
        commands.append("\n".join(lines[i:]))
        return [command.strip() for command in commands if command]

    @classmethod
    def _rewrite_init_command(cls, command: str) -> str:
        """
        SQLite init scripts can contain dot commands, which can only be executed
        by the CLI, not the python API. Here, we rewrite some common ones into
        SQL, and rewrite the others to be no-ops.
        """
        if not command.startswith("."):
            return command
        elif command.startswith(".open"):
            return cls._rewrite_dot_open(command)
        elif command.startswith(".load"):
            return cls._rewrite_dot_load(command)
        else:
            return ""

    @staticmethod
    def _rewrite_dot_open(command: str) -> str:
        """
        Rewrites .open command into its approx SQL equivalent.
        (This attaches databases instead of opening a new connection).

        Options (--readonly, etc.) are not supported.
        """
        args = command.split()[1:]
        if not args:
            return "attach ''"
        elif args[-1] == ":memory:":
            return "attach ':memory:';"
        else:
            # options are not supported.
            db_path = Path(args[-1])
            return f"attach '{db_path}' as {db_path.stem};"

    @staticmethod
    def _rewrite_dot_load(command: str) -> str:
        """
        Rewrites .load command into its SQL equivalent, load_extension().
        """
        args = command.split()[1:]
        if not args:
            raise sqlite3.ProgrammingError("Could not execute .load with no args.")
        elif len(args) == 1:
            return f"select load_extension('{args[0]}');"
        elif len(args) == 2:
            return f"select load_extension('{args[0]}', '{args[1]}');"
        else:
            raise sqlite3.ProgrammingError(
                f"Could not execute .load with following args: {args}"
            )



from __future__ import annotations

import sys
from pathlib import Path
from typing import Awaitable, Callable

import pytest
from textual.widgets import Input

from harlequin import HarlequinKeys

USER_CONFIG_PATH = Path("/tmp") / "harlequin"


@pytest.fixture(autouse=True)
def mock_config_loader(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "harlequin.keys_app.get_config_for_profile", lambda **_: (dict(), [])
    )
    monkeypatch.setattr(
        "harlequin.keys_app.get_highest_priority_existing_config_file", lambda **_: None
    )
    monkeypatch.setattr(
        "harlequin.keys_app.user_config_path", lambda **_: USER_CONFIG_PATH
    )


@pytest.fixture
def keys_app() -> HarlequinKeys:
    return HarlequinKeys()


@pytest.mark.asyncio
async def test_keys_app(
    keys_app: HarlequinKeys,
    app_snapshot: Callable[..., Awaitable[bool]],
) -> None:
    target_path = USER_CONFIG_PATH / "config.toml"
    target_path.unlink(missing_ok=True)
    app = keys_app
    snap_results: list[bool] = []
    async with app.run_test(size=(120, 36)) as pilot:
        while (
            app.active_keymap_names is None
            or app.bindings is None
            or app.unmodifed_bindings is None
            or app.table is None
        ):
            await pilot.pause()
        snap_results.append(await app_snapshot(app, "Initialization"))

        await pilot.press("down", "down", "down", "down", "enter")
        await pilot.pause()
        snap_results.append(await app_snapshot(app, "Edit Modal"))

        await pilot.press("enter")
        await pilot.pause()
        snap_results.append(await app_snapshot(app, "Enter Key Modal"))

        await pilot.press("f3")
        await pilot.pause(0.1)
        await pilot.wait_for_animation()
        snap_results.append(await app_snapshot(app, "Edit Modal: f3"))

        await pilot.press("tab", "tab", "enter", "f4")
        await pilot.pause(0.1)
        await pilot.wait_for_animation()
        snap_results.append(await app_snapshot(app, "Edit Modal: f3 and f4"))

        await pilot.press("shift+tab", "enter")
        await pilot.pause()
        await pilot.wait_for_animation()
        snap_results.append(await app_snapshot(app, "Edit Modal: f4 removed"))

        await pilot.press("tab", "tab", "tab", "tab", "enter")
        await pilot.pause()
        await pilot.wait_for_animation()
        snap_results.append(await app_snapshot(app, "Main modal, Focus F3"))

        await pilot.press("ctrl+q")
        await pilot.pause()
        # the quit modal should now be visible. We make some tweaks so tests
        # pass consistently
        input_widgets = app.query(Input)
        for widget in input_widgets:
            widget.cursor_blink = False  # prevent flaky tests
        path_input = app.query_one("#path_input", expect_type=Input)
        assert path_input.value == str(target_path)
        if sys.platform == "win32":
            # on windows the path will use backslashes, which makes
            # the snapshot fail. Use forward slashes instead.
            path_input.value = target_path.as_posix()
        await pilot.wait_for_animation()
        snap_results.append(await app_snapshot(app, "Quit modal"))

        await pilot.press("f", "o", "o", "tab", "tab", "tab", "enter")
        assert app.return_code == 0
        assert target_path.exists()

        assert all(snap_results)



from typing import Union

import pyperclip
from rich.markup import escape
from textual import events
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Static


class ClickableStatic(Static):
    def on_click(self, message: events.Click) -> None:
        message.stop()
        try:
            pyperclip.copy(str(self.renderable))
        except pyperclip.PyperclipException:
            pass
        else:
            self.app.notify("Error copied to clipboard.")


class ErrorModal(ModalScreen):
    def __init__(
        self,
        title: str,
        header: str,
        error: BaseException,
        name: Union[str, None] = None,
        id: Union[str, None] = None,  # noqa: A002
        classes: Union[str, None] = None,
    ) -> None:
        super().__init__(name, id, classes)
        self.title = title
        self.header = header
        self.error = error

    def compose(self) -> ComposeResult:
        with Vertical(id="error_outer"):
            yield Static(self.header, id="error_header")
            with Vertical(id="error_inner"):
                with VerticalScroll():
                    yield ClickableStatic(escape(str(self.error)), id="error_info")
            yield Static(
                "Press any key to continue. Click error to copy.", id="error_footer"
            )

    def on_mount(self) -> None:
        container = self.query_one("#error_outer")
        container.border_title = self.title

    def on_key(self) -> None:
        self.app.pop_screen()

    def on_click(self, message: events.Click) -> None:
        self.app.pop_screen()



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import Awaitable, Callable

import pytest

from harlequin import Harlequin, HarlequinAdapter
from harlequin.config import get_config_for_profile

QUERY = dedent(
    """
    select *
    from
        (
            values
                (1, 2, 3),
                (4, 5, 6),
                (7, 8, 9),
                (10, 11, 12),
                (13, 14, 15),
                (16, 17, 18),
                (19, 20, 21)
        ) foo(a, b, c)
"""
).strip()


@pytest.mark.asyncio
async def test_results_viewer_bindings(
    duckdb_adapter: type[HarlequinAdapter],
    data_dir: Path,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    config_path = (
        data_dir / "functional_tests" / "test_keymap_from_config" / "config.toml"
    )
    profile, my_keymaps = get_config_for_profile(
        config_path=config_path, profile_name=None
    )
    app = Harlequin(
        duckdb_adapter([":memory:"], no_init=True),
        keymap_names=profile["keymap_name"],
        user_defined_keymaps=my_keymaps,
    )
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()

        q = QUERY
        app.editor.text = q
        await pilot.press("ctrl+j")

        while (table := app.results_viewer.get_visible_table()) is None:
            await pilot.pause()

        assert table is not None
        assert table.cursor_coordinate == (0, 0)
        assert table.selection_anchor_coordinate is None

        # simple navigation
        await pilot.press("s")
        assert table.cursor_coordinate == (1, 0)
        assert table.selection_anchor_coordinate is None
        await pilot.press("d")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate is None
        await pilot.press("d")
        assert table.cursor_coordinate == (1, 2)
        assert table.selection_anchor_coordinate is None
        await pilot.press("a")
        assert table.cursor_coordinate == (1, 1)
        assert table.selection_anchor_coordinate is None
        await pilot.press("w")
        assert table.cursor_coordinate == (0, 1)
        assert table.selection_anchor_coordinate is None



from __future__ import annotations

import sys
from typing import Awaitable, Callable

import pytest
from textual.message import Message

from harlequin import Harlequin


def transaction_button_visible(app: Harlequin) -> bool:
    """
    Skip snapshot checks for versions of that app showing the autocommit button.
    """
    return sys.version_info >= (3, 12) and "Sqlite" in app.adapter.__class__.__name__


@pytest.mark.asyncio
async def test_run_query_bar(
    app_all_adapters_small_db: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters_small_db
    snap_results: list[bool] = []
    messages: list[Message] = []
    async with app.run_test(size=(120, 36), message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        # initialization
        bar = app.run_query_bar
        assert bar.limit_checkbox.value is False
        assert bar.limit_input.value == "500"
        assert bar.limit_value is None

        # query without any limit by clicking the button;
        # dataset has 857 records
        assert app.editor is not None
        app.editor.text = "select * from drivers"
        await pilot.click("#run_query")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        table = app.results_viewer.get_visible_table()
        assert table
        assert table.row_count == table.source_row_count == 857
        snap_results.append(await app_snapshot(app, "No limit"))

        # apply a limit by clicking the limit checkbox
        await pilot.click(bar.limit_checkbox.__class__)
        assert bar.limit_checkbox.value is True
        assert bar.limit_value == 500
        await pilot.click("#run_query")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        table = app.results_viewer.get_visible_table()
        assert table
        assert table.row_count == table.source_row_count == 500
        snap_results.append(await app_snapshot(app, "Limit 500"))

        # type an invalid limit, checkbox should be unchecked
        # and a tooltip should appear on hover
        await pilot.click(bar.limit_input.__class__)
        await pilot.press("a")
        assert bar.limit_input.value == "a500"
        assert bar.limit_value is None
        assert bar.limit_checkbox.value is False
        assert bar.limit_input.tooltip is not None
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "Invalid limit"))

        # type a valid limit
        await pilot.press("backspace")
        await pilot.press("delete")
        await pilot.press("1")
        assert bar.limit_input.value == "100"
        assert bar.limit_value == 100
        assert bar.limit_checkbox.value is True
        assert bar.limit_input.tooltip is None

        # run the query with a smaller limit
        await pilot.click("#run_query")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        table = app.results_viewer.get_visible_table()
        assert table
        assert table.row_count == table.source_row_count == 100
        snap_results.append(await app_snapshot(app, "Limit 100"))

        if not transaction_button_visible(app):
            assert all(snap_results)


@pytest.mark.skipif(
    sys.version_info < (3, 12), reason="SQLite in Python < 3.12 won't show txn button"
)
@pytest.mark.asyncio
async def test_transaction_button(
    app_small_sqlite: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_small_sqlite
    snap_results: list[bool] = []
    async with app.run_test(size=(120, 36)) as pilot:
        await wait_for_workers(app)
        while app.editor is None or app.connection is None:
            await pilot.pause()

        assert app.connection.transaction_mode
        assert app.connection.transaction_mode.label == "Auto"
        snap_results.append(await app_snapshot(app, "Initialize with Tx: Auto"))
        await pilot.click("#transaction_button")
        await pilot.pause(0.3)
        await pilot.wait_for_animation()
        assert app.connection.transaction_mode
        assert app.connection.transaction_mode.label == "Manual"
        assert app.connection.transaction_mode.commit is not None
        assert app.connection.transaction_mode.rollback is not None
        snap_results.append(await app_snapshot(app, "After click with Tx: Manual"))

        assert all(snap_results)



import asyncio

from textual.widgets.text_area import Selection
from textual.widgets.tree import TreeNode

from harlequin import Harlequin
from harlequin.catalog import CatalogItem
from harlequin.colors import VALID_THEMES
from harlequin_duckdb import DuckDbAdapter

TEXT = """
select
    drivers.surname,
    drivers.forename,
    drivers.nationality,
    avg(driver_standings.position) as avg_standing,
    avg(driver_standings.points) as avg_points
from driver_standings
join drivers on driver_standings.driverid = drivers.driverid
join races on driver_standings.raceid = races.raceid
group by 1, 2, 3
order by avg_standing asc
""".strip()


async def wait_for_filtered_workers(app: Harlequin) -> None:
    filtered_workers = [
        w for w in app.workers if w.name != "_database_tree_background_loader"
    ]
    if filtered_workers:
        await app.workers.wait_for_complete(filtered_workers)


async def save_all_screenshots() -> None:
    adapter = DuckDbAdapter(("f1.db",), no_init=True)
    for theme in VALID_THEMES:
        print(f"Screenshotting {theme}")
        app = Harlequin(adapter=adapter, theme=theme)
        async with app.run_test(size=(120, 36)) as pilot:
            await wait_for_filtered_workers(app)
            await pilot.pause()
            if app.editor is None:
                await pilot.pause(0.2)
            assert app.editor is not None
            app.editor.text = TEXT
            app.editor.selection = Selection((9, 0), (9, 16))

            async def _expand_and_wait(node: TreeNode[CatalogItem]) -> None:
                node.expand()
                while not node.children:
                    if getattr(node.data, "loaded", True):
                        break
                    await pilot.pause()

            app.data_catalog.database_tree.root.expand()
            for db_node in app.data_catalog.database_tree.root.children:
                await _expand_and_wait(db_node)
                for schema_node in db_node.children:
                    await _expand_and_wait(schema_node)
                    for table_node in schema_node.children:
                        if str(table_node.label).startswith("drivers"):
                            await _expand_and_wait(table_node)
            app.data_catalog.database_tree.cursor_line = 7
            await pilot.press("ctrl+j")
            app.save_screenshot(filename=f"{theme}.svg", path="./static/themes/")
        app.exit()


asyncio.run(save_all_screenshots())



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar, Generic, Protocol, Sequence, TypeVar

from textual.message import Message

if TYPE_CHECKING:
    from harlequin.adapter import HarlequinConnection
    from harlequin.driver import HarlequinDriver


@dataclass
class CatalogItem:
    """
    A basic representation of a database object for Harlequin's Data Catalog.

    Args:
        qualified_identifier (str): The fully-scoped and (optionally) quoted database
            identifier for this object. Should be globally unique for the Connection.
            e.g., a postgres table: "mydb"."myschema"."mytable"
        query_name (str): The text to be inserted into the query editor for this
            object. This is typically the qualified_indentifier, but often
            for columns it is just the quoted name of the column (without
            the qualifying schema/table/etc).
        label (str): The unqualified name for this object, to appear in the Data
            Catalog widget.
        type_label (str): A short (1-3 chars) label denoting the type of this object
            or the data it contains. e.g., ("t") for "table" or "##" for
            "bigint column".
        children (list[CatalogItem]): Other nodes nested under this one (e.g.,
            a list of columns if this is a table.). If the list is empty and
            a CatalogItem subclass implements the `fetch_children()` method,
            Harlequin will attempt to call that method to lazy-load children.
    """

    qualified_identifier: str
    query_name: str
    label: str
    type_label: str
    children: list["CatalogItem"] = field(default_factory=list)


TCatalogItem_contra = TypeVar(
    "TCatalogItem_contra", bound=CatalogItem, contravariant=True
)
TAdapterConnection_contra = TypeVar(
    "TAdapterConnection_contra", bound="HarlequinConnection", contravariant=True
)


class Interaction(Protocol[TCatalogItem_contra]):
    def __call__(
        self,
        item: TCatalogItem_contra,
        driver: "HarlequinDriver",
    ) -> None: ...


@dataclass
class InteractiveCatalogItem(CatalogItem, Generic[TAdapterConnection_contra]):
    """
    An advanced representation of a database object that can lazy-load its
    children and provide other interactions via a context menu. Subclass
    this class and define the INTERACTIONS class variable to populate
    Harlequin's context menu.

    Each list item is a tuple with a label (for the context menu) and a
    callable that takes three arguments: the CatalogItem node that is clicked
    on, the initialized HarlequinConnection, and a HarlequinDriver that
    provides a simplified interface to the Harlequin App and allows the
    interactions to do things like insert text into the editor, show
    a confirmation modal, and display a notification (toast).
    """

    INTERACTIONS: ClassVar[Sequence[tuple[str, Interaction]] | None] = None
    loaded: bool = False
    """
    Harlequin will set loaded to True after calling fetch_children. You
    can prevent calls to fetch_children by initializing loaded=True.
    """
    connection: TAdapterConnection_contra | None = None

    def fetch_children(self) -> Sequence[CatalogItem]:
        """
        Returns a list of CatalogItems (or subclass instances, like other
        InteractiveCatalogItems) to be shown under this item in the catalog
        tree viewer. Return an empty list if this item has no children.
        """
        return []


@dataclass
class Catalog:
    """
    A representation of the queryable objects in a database, to be displayed in the
    DataCatalog widget.

    Args:
        items (list[CatalogItem]): A list of CatalogItem nodes, which each represent
            a database object, like a database, schema, table, or column.
    """

    items: list[CatalogItem]


class NewCatalog(Message):
    def __init__(self, catalog: Catalog) -> None:
        self.catalog = catalog
        super().__init__()


class NewCatalogItems(Message):
    def __init__(self, parent: CatalogItem, items: list[CatalogItem]) -> None:
        self.parent = parent
        self.items = items
        super().__init__()



from __future__ import annotations

import sys
from typing import Awaitable, Callable

import pytest
from textual.message import Message

from harlequin import Harlequin
from harlequin.app import QueriesExecuted, QuerySubmitted, ResultsFetched
from harlequin.components import ErrorModal


def transaction_button_visible(app: Harlequin) -> bool:
    """
    Skip snapshot checks for versions of that app showing the autocommit button.
    """
    return sys.version_info >= (3, 12) and "Sqlite" in app.adapter.__class__.__name__


@pytest.mark.asyncio
async def test_select_1(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        assert app.title == "Harlequin"
        assert app.focused.__class__.__name__ == "TextAreaPlus"

        q = "select 1 as foo"
        for key in q:
            await pilot.press(key)
        await pilot.press("ctrl+j")  # alias for ctrl+enter

        await pilot.pause()
        [query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == [q]
        await wait_for_workers(app)
        await pilot.pause()
        [query_executed_message] = [
            m for m in messages if isinstance(m, QueriesExecuted)
        ]
        assert query_executed_message.query_count == 1
        assert query_executed_message.cursors
        await wait_for_workers(app)
        await pilot.pause()
        [results_fetched_message] = [
            m for m in messages if isinstance(m, ResultsFetched)
        ]
        assert results_fetched_message.errors == []
        table = app.results_viewer.get_visible_table()
        assert table
        assert table.source_row_count == table.row_count == 1
        # sqlite on py3.12 will show the Tx: Auto button, and snap
        # will fail
        if not transaction_button_visible(app):
            assert await app_snapshot(app, "select 1 as foo")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "query",
    [
        "select 1+1",
        "select 'a' as foo",
        "select null",
        "select null as foo",
        "",
        "select 1 where false",
        "select 1 union all select 'hi'",
        "select 'hi' union all select 1",
        """select '[/foo]' as "[/bar]" """.strip(),
    ],
)
async def test_queries_do_not_crash_all_adapters(
    app_all_adapters: Harlequin,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
    query: str,
) -> None:
    app = app_all_adapters
    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+j")
        await pilot.pause()

        if query:
            [query_submitted_message] = [
                m for m in messages if isinstance(m, QuerySubmitted)
            ]
            assert query_submitted_message.queries == [query]
            await wait_for_workers(app)
            await pilot.pause()
            [query_executed_message] = [
                m for m in messages if isinstance(m, QueriesExecuted)
            ]
            assert query_executed_message.cursors
        if query and query != "select 1 where false":
            await pilot.pause()
            await wait_for_workers(app)
            table = app.results_viewer.get_visible_table()
            assert table is not None
            assert table.row_count >= 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "query",
    [
        "SELECT {'x': 1, 'y': 2, 'z': 3}",  # struct
        # also a struct:
        "SELECT {'yes': 'duck', 'maybe': 'goose', 'huh': NULL, 'no': 'heron'}",
        "SELECT {'key1': 'string', 'key2': 1, 'key3': 12.345}",  # struct
        """SELECT {'birds':
            {'yes': 'duck', 'maybe': 'goose', 'huh': NULL, 'no': 'heron'},
        'aliens':
            NULL} as bar""",  # struct
        "select {'a': 5} union all select {'a': 6}",  # struct
        "select map {'a': 5}",  # map
        "select map {'a': 5} union all select map {'b': 6}",  # map
        "SELECT map { 1: 42.001, 5: -32.1 }",  # map
        "SELECT map { ['a', 'b']: [1.1, 2.2], ['c', 'd']: [3.3, 4.4] }",  # map
        "SELECT [1, 2, 3]",  # list
        "SELECT ['duck', 'goose', NULL, 'heron'];",  # list
        "SELECT [['duck', 'goose', 'heron'], NULL, ['frog', 'toad'], []];",  # list
        "set timezone = 'America/New_York'; select '2024-01-01'::timestamptz;",
        "set timezone = 'America/New_York'; select '1-1-1T00:00:00Z'::timestamptz",
    ],
)
async def test_queries_do_not_crash(
    app: Harlequin,
    query: str,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = query
        await pilot.press("ctrl+a")
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        table = app.results_viewer.get_visible_table()
        assert table is not None
        assert table.row_count >= 1


@pytest.mark.asyncio
async def test_multiple_queries(
    app_all_adapters: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    snap_results: list[bool] = []
    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        q = "select 1; select 2"
        app.editor.text = q
        await pilot.press("ctrl+j")

        # should only run one query
        await wait_for_workers(app)
        await pilot.pause()
        [query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == ["select 1;"]
        table = app.results_viewer.get_visible_table()
        assert table
        assert table.row_count == table.source_row_count == 1
        assert "hide-tabs" in app.results_viewer.classes
        await wait_for_workers(app)
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "One query"))

        app.editor.focus()
        await pilot.press("ctrl+a")
        await pilot.press("ctrl+j")
        # should run both queries
        await wait_for_workers(app)
        await pilot.pause()
        await pilot.wait_for_scheduled_animations()
        [_, query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == ["select 1;", "select 2"]
        assert app.results_viewer.tab_count == 2
        assert "hide-tabs" not in app.results_viewer.classes
        await wait_for_workers(app)
        await pilot.pause(0.5)
        await pilot.wait_for_scheduled_animations()
        snap_results.append(await app_snapshot(app, "Both queries"))
        assert app.results_viewer.active == "result-1"
        await pilot.press("k")
        await pilot.wait_for_scheduled_animations()
        assert app.results_viewer.active == "result-2"
        snap_results.append(await app_snapshot(app, "Both queries, tab 2"))
        await pilot.press("k")
        await pilot.wait_for_scheduled_animations()
        assert app.results_viewer.active == "result-1"
        snap_results.append(await app_snapshot(app, "Both queries, tab 1"))
        await pilot.press("j")
        assert app.results_viewer.active == "result-2"
        await pilot.press("j")
        assert app.results_viewer.active == "result-1"

        if not transaction_button_visible(app):
            assert all(snap_results)


@pytest.mark.asyncio
async def test_single_query_terminated_with_semicolon(
    app_all_adapters: Harlequin,
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    messages: list[Message] = []
    async with app.run_test(message_hook=messages.append) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        q = "select 1;    \n\t\n"
        app.editor.text = q
        await pilot.press("ctrl+j")

        # should only run current query
        await wait_for_workers(app)
        await pilot.pause()
        [query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == ["select 1;"]
        assert app.results_viewer.tab_count == 1

        app.editor.focus()
        await pilot.press("ctrl+a")
        await pilot.press("ctrl+j")

        # should not run whitespace query, even though included
        # in selection.
        await wait_for_workers(app)
        await pilot.pause()
        [_, query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == ["select 1;"]
        assert app.results_viewer.tab_count == 1

        app.editor.focus()
        await pilot.press("ctrl+end")
        await pilot.press("ctrl+j")
        # should run previous query
        await wait_for_workers(app)
        await pilot.pause()
        [*_, query_submitted_message] = [
            m for m in messages if isinstance(m, QuerySubmitted)
        ]
        assert query_submitted_message.queries == ["select 1;"]
        assert app.results_viewer.tab_count == 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "bad_query",
    [
        "select",  # errors when building cursor
        "select 0::struct(id int)",  # errors when fetching data
        "select; select 0::struct(id int)",  # multiple errors
        "select 1; select 0::struct(id int)",  # one error, mult queries
        "select 0::struct(id int); select 1",  # one error, mult queries, err first
    ],
)
async def test_query_errors(
    app_all_adapters: Harlequin,
    bad_query: str,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    app = app_all_adapters
    snap_results: list[bool] = []
    async with app.run_test(size=(120, 36)) as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause()
        app.editor.text = bad_query

        await pilot.press("ctrl+a")
        await pilot.press("ctrl+j")
        await wait_for_workers(app)
        await pilot.pause()
        assert len(app.screen_stack) == 2
        assert isinstance(app.screen, ErrorModal)
        snap_results.append(await app_snapshot(app, "Error visible"))

        await pilot.press("space")
        assert len(app.screen_stack) == 1

        # data table and query bar should be responsive
        assert "non-responsive" not in app.run_query_bar.classes
        assert "non-responsive" not in app.results_viewer.classes
        snap_results.append(await app_snapshot(app, "After dismissing error"))

        if not transaction_button_visible(app):
            assert all(snap_results)


@pytest.mark.asyncio
async def test_rich_markup(
    app: Harlequin,
    app_snapshot: Callable[..., Awaitable[bool]],
    wait_for_workers: Callable[[Harlequin], Awaitable[None]],
) -> None:
    async with app.run_test() as pilot:
        await wait_for_workers(app)
        while app.editor is None:
            await pilot.pause(0.1)

        q = "select '[some text]', '[red]some text[/]'"
        app.editor.text = q
        await pilot.press("ctrl+j")  # alias for ctrl+enter

        await pilot.pause()
        await wait_for_workers(app)
        await pilot.pause()
        assert await app_snapshot(app, "select markup")



from harlequin.autocomplete.completers import (
    MemberCompleter,
    WordCompleter,
    completer_factory,
)
from harlequin.autocomplete.completion import HarlequinCompletion

__all__ = [
    "HarlequinCompletion",
    "MemberCompleter",
    "WordCompleter",
    "completer_factory",
]



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING, Literal, Sequence

from harlequin.catalog import CatalogItem
from harlequin.exception import HarlequinQueryError

if TYPE_CHECKING:
    from harlequin.driver import HarlequinDriver
    from harlequin_sqlite.catalog import (
        ColumnCatalogItem,
        RelationCatalogItem,
    )


def execute_drop_relation_statement(
    item: "RelationCatalogItem",
    driver: "HarlequinDriver",
    relation_type: Literal["view", "table"],
) -> None:
    def _drop_relation() -> None:
        if item.connection is None:
            return
        try:
            item.connection.execute(f"drop {relation_type} {item.qualified_identifier}")
        except HarlequinQueryError:
            driver.notify(
                f"Could not drop {relation_type} {item.label}", severity="error"
            )
            raise
        else:
            driver.notify(f"Dropped {relation_type} {item.label}")
            driver.refresh_catalog()

    driver.confirm_and_execute(callback=_drop_relation)


def execute_drop_table_statement(
    item: "RelationCatalogItem", driver: "HarlequinDriver"
) -> None:
    execute_drop_relation_statement(item=item, driver=driver, relation_type="table")


def execute_drop_view_statement(
    item: "RelationCatalogItem", driver: "HarlequinDriver"
) -> None:
    execute_drop_relation_statement(item=item, driver=driver, relation_type="view")


def show_describe_relation(
    item: "RelationCatalogItem",
    driver: "HarlequinDriver",
) -> None:
    driver.insert_text_in_new_buffer(
        text=f"pragma table_info({item.qualified_identifier})"
    )


def show_relation_ddl(
    item: "RelationCatalogItem",
    driver: "HarlequinDriver",
) -> None:
    if item.parent is None or item.connection is None:
        return
    cursor = item.connection.execute(
        f"""
        select sql
        from sqlite_schema
        where tbl_name = '{item.label}'
        limit 1
        """
    )
    assert cursor is not None
    result = cursor.fetchone()
    if result is not None:
        driver.insert_text_in_new_buffer(text=result[0])


def show_select_star(
    item: "RelationCatalogItem",
    driver: "HarlequinDriver",
) -> None:
    driver.insert_text_in_new_buffer(
        dedent(
            f"""
            select *
            from {item.qualified_identifier}
            limit 100
            """.strip("\n")
        )
    )


def insert_columns_at_cursor(
    item: "RelationCatalogItem",
    driver: "HarlequinDriver",
) -> None:
    if item.loaded:
        cols: Sequence["CatalogItem" | "ColumnCatalogItem"] = item.children
    else:
        cols = item.fetch_children()
    driver.insert_text_at_selection(text=",\n".join(c.query_name for c in cols))



import asyncio
from unittest.mock import patch

from textual.widgets.text_area import Selection

from harlequin import Harlequin
from harlequin.editor_cache import BufferState, Cache
from harlequin_duckdb import DuckDbAdapter


async def wait_for_filtered_workers(app: Harlequin) -> None:
    filtered_workers = [
        w for w in app.workers if w.name != "_database_tree_background_loader"
    ]
    if filtered_workers:
        await app.workers.wait_for_complete(filtered_workers)


async def load_lots_of_buffers() -> None:
    with patch("harlequin.components.code_editor.load_cache") as mock_load_cache:
        buff = BufferState(
            selection=Selection((0, 0), (0, 0)),
            text="select 1; " * 20,
        )
        cache = Cache(focus_index=0, buffers=[buff] * 10)
        mock_load_cache.return_value = cache

        adapter = DuckDbAdapter((":memory:",), no_init=True)
        app = Harlequin(adapter=adapter)

        async with app.run_test() as pilot:
            await wait_for_filtered_workers(app)
            await pilot.pause()
        app.exit()


asyncio.run(load_lots_of_buffers())
print("Ran successfully")



from harlequin.adapter import HarlequinAdapter, HarlequinConnection, HarlequinCursor
from harlequin.app import Harlequin
from harlequin.autocomplete import HarlequinCompletion
from harlequin.keymap import HarlequinKeyBinding, HarlequinKeyMap
from harlequin.keys_app import HarlequinKeys
from harlequin.options import HarlequinAdapterOption, HarlequinCopyFormat
from harlequin.transaction_mode import HarlequinTransactionMode

__all__ = [
    "Harlequin",
    "HarlequinAdapter",
    "HarlequinAdapterOption",
    "HarlequinCompletion",
    "HarlequinConnection",
    "HarlequinCopyFormat",
    "HarlequinCursor",
    "HarlequinTransactionMode",
    "HarlequinKeys",
    "HarlequinKeyMap",
    "HarlequinKeyBinding",
]
